from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
# from Cython.Build import cythonize

setup(
  name = "mc",
  cmdclass = {"build_ext": build_ext},
  ext_modules =
  [
    Extension("mc",
              ["mc.pyx"],
              extra_compile_args = ["-fopenmp"],
              extra_link_args=['-fopenmp']
              )
  ]
)



#! /usr/bin/python

"""
A python API and CLI to perform pKa calculations on peptides,
proteins or lipid bilayers.
"""

import config
from cli import checkParsedInput, readSettings, inputParametersFilter
from cleaning import inputPDBCheck, cleanPDB
from formats import convertTermini, pdb2gro
import log

from delphi4py.delphi4py import DelPhi4py

from molecule import Molecule
from concurrency import startPoolProcesses, runDelPhiSims
import os

__author__ = "Pedro Reis"
__version__ = "0.4"

__email__ = "pdreis@fc.ul.pt"
__status__ = "Development"


def getTitrableSites(pdb):
    """Gets the all titrable sites from the pdb

    Returns an input structure to the Titration class containing all
    titrable sites found in the pdb file. 

    Args:
        pdb (str): The filename a PDB file
    
    Returns: 
        A dict mapping all titrable sites found in the pdb file

    """
    config.f_in = pdb
    config.params['ffID'] = 'G54A7'

    tit_mol = Molecule()

    tit_mol.makeSimpleSites()

    sites = tit_mol.getSitesOrdered()
    sites_keys = []
    for site in sites:
        sitename = site.getName()
        sitenumber = site.getResNumber()
        sitenumber = convertTermini(sitenumber)
        if sitename in ('NTR', 'CTR'):
            if sitename == 'NTR':
                sites_keys.append('{0}N'.format(sitenumber))
            else:
                sites_keys.append('{0}C'.format(sitenumber))
        else:
            sites_keys.append(str(sitenumber))

    return sites_keys


class Titration(object):
    """Main pypka class

    Serves as a wrapper to all other classes

    Args:
        parameters
        sites
        debug
        datfile

    Attributes:
        _pKas: A dict with the calculated pKa values.

    """
    def __init__(self, parameters, sites='all', debug=None, datfile=None):
        """
        self._pKas = {}
        """
        self._pKas = {}

        if sites:  # None if from CLI
            parameters['sites_A'] = sites
            if sites != 'all':
                config.sites = sites
            else:
                config.sites = {}

        if datfile:
            config.f_dat = datfile
            parameters = readSettings(datfile) 

        config.debug = debug

        # Check Input Variables Validity
        inputParametersFilter(parameters)

        print('Start Preprocessing')
        self.preprocessing()
        self.processDelPhiParams()

        print('Start PB Calculations')

        self.DelPhiLaunch()
        print('API exited successfully')

    def preprocessing(self):
        # Creating instance of TitratingMolecule
        config.tit_mole = Molecule()

        if config.f_in_extension == 'gro':
            groname = config.f_in
            if len(config.sites) > 0:
                chains_length, chains_res = inputPDBCheck(config.f_in, config.sites)
                config.tit_mole.loadSites(chains_length, chains_res)
            else:
                log.inputVariableError('sites',
                                       'defined.',
                                       'When using a .gro file format input is used, '
                                       'sites needs to be defined')
        if config.f_in_extension == 'pdb':
            # Reading .st files
            # If the titrable residues are defined
            if len(config.sites) > 0:
                chains_length, chains_res = inputPDBCheck(config.f_in,
                                                          config.sites)
                config.tit_mole.loadSites(chains_length, chains_res)
            # If the titrable residues are not defined and
            # the input pdb file is incomplete
            elif config.params['clean_pdb']:
                chains_res, sites = config.tit_mole.makeSimpleSites()
            # If the titrable residues are not defined and
            # the input pdb is not missing any atoms
            else:
                chains_res, sites = config.tit_mole.makeSimpleSites()
                config.tit_mole.deleteAllSites()
                config.tit_mole.makeSites(sites=sites)

            # Creates a .pdb input for DelPhi
            # where residues are in their standard state
            if config.params['clean_pdb']:
                inputpqr = 'clean.pqr'
                outputpqr = 'cleaned_tau.pqr'
                sites = config.tit_mole.getSites()
                site_numb_n_ref = {}
                for site in sites:
                    site_numb_n_ref[site] = sites[site].getRefTautomerName()
                
                cleanPDB(config.f_in, config.pdb2pqr, chains_res,
                         inputpqr, outputpqr, site_numb_n_ref)
                config.tit_mole.deleteAllSites()
                config.tit_mole.makeSites(useTMPgro=True, sites=list(sites.keys()))
            else:
                pdb2gro(config.f_in, 'TMP.gro', config.tit_mole.box,
                        config.sites)
            groname = 'TMP.gro'

        config.tit_mole.readGROFile(groname)
        if not config.debug and config.f_in_extension == 'pdb' and os.path.isfile('TMP.gro'):
            os.remove('TMP.gro')

    def processDelPhiParams(self):
        # Storing DelPhi parameters and Creates DelPhi data structures
        if config.debug:
            logfile = 'LOG_readFiles'
        else:
            logfile = config.f_log

        delphimol = DelPhi4py(config.f_crg, config.f_siz, 'delphi_in_stmod.pdb',
                           config.tit_mole.getNAtoms(),
                           config.params['gsize'],
                           config.params['scaleM'],
                           config.params['precision'],
                           epsin=config.params['epsin'],
                           epsout=config.params['epssol'],
                           conc=config.params['ionicstr'],
                           ibctyp=config.params['bndcon'],
                           res2=config.params['maxc'],
                           nlit=config.params['nlit'],
                           nonit=config.params['nonit'],
                           relfac=0.0,
                           relpar=0.0,
                           pbx=config.params['pbx'],
                           pby=config.params['pby'],
                           isurftype=config.nanoshaper,
                           debug=config.debug, outputfile=logfile)
        if not config.debug:
            os.remove('delphi_in_stmod.pdb')

        log.checkDelPhiErrors(logfile, 'readFiles')
        
        # Loads delphi4py object as TitratingMolecule attributes
        config.tit_mole.loadDelPhiParams(delphimol)

        if config.debug:
            config.tit_mole.printAllSites()
            config.tit_mole.printAllTautomers()
            print(delphimol)

    def DelPhiLaunch(self):
        if len(config.tit_mole.getSites()) < 1:
            raise Exception('At least one site has to be correctly defined.')

        # Runs DelPhi simulations for all tautomers
        results = startPoolProcesses(runDelPhiSims,
                                     config.tit_mole.iterAllSitesTautomers(),
                                     config.params['ncpus'])

        # Calculates the pKint of all tautomers
        config.tit_mole.calcpKint(results)

        # Calculates sites interaction energies and write .dat file
        config.tit_mole.calcSiteInteractionsParallel(config.params['ncpus'])

        #  Monte Carlo sampling
        pKas, pmeans = config.tit_mole.runMC()

        sites = config.tit_mole.getSitesOrdered()

        c = -1
        for i in pKas:
            c += 1
            site = sites[c]._res_number
            pK = i[0]
            if pK != 100.0:
                self._pKas[site] = pK
            else:
                self._pKas[site] = '-'
        self._pmeans = pmeans

    def getAverageProt(self, site, pH):
        """Calculates the average protonation of a site at a given pH value

        Args:
            site (str): Residue number of the site with the suffix 'N'
            or 'C' to indicate a N-ter or C-ter, respectively.

            pH (float): pH value 
        
        Returns:  
            A float of the average protonation of the site at the
            selected pH value
        """
        pKa = self[site]
        if isinstance(pKa, str):
            return 'pk Not In Range'
        average_prot = 10 ** (pKa - pH) / (1 + 10 ** (pKa - pH))
        return average_prot

    def getProtState(self, site, pH):
        """Indicates the most probable protonation state of a site at a given
        pH value

        Args:
            site (str): Residue number of the site with the suffix 'N'
            or 'C' to indicate a N-ter or C-ter, respectively.

            pH (float): pH value 

        Returns: 

            A tuple with two elements. Example: ('protonated', 0.9)

            The first element is a string indicating the most probable
            state of the site. It can be either 'protonated',
            'deprotonated' or 'undefined'. The 'undefined' state is
            prompted if the average protonation state is between 0.1 and 0.9.
          
            The second element is a float of the average protonation of the site

        """
        state = 'undefined'
        average_prot = self.getAverageProt(site, pH)

        if isinstance(average_prot, str):
            return state, average_prot

        if average_prot > 0.9:
            state = 'protonated'
        elif average_prot < 0.1:
            state = 'deprotonated'

        return state, average_prot

    def __iter__(self):
        self._iterpKas = []
        for site in config.tit_mole.getSitesOrdered():
            self._iterpKas.append(site.getResNumber())
        self._iternumb = -1
        self._itermax = len(self._iterpKas)
        return self

    def __next__(self):
        self._iternumb += 1
        if self._iternumb < self._itermax:
            site = self._iterpKas[self._iternumb]
            if site > config.terminal_offset:
                if site - config.terminal_offset == config.tit_mole._NTR:
                    site = 'NTR'
                elif site - config.terminal_offset == config.tit_mole._CTR:
                    site = 'CTR'
                else:
                    raise Exception('Something is terribly wrong')
            return site
        else:
            raise StopIteration

    def correct_site_numb(self, numb):
        if numb == 'NTR':
            numb = config.tit_mole._NTR + config.terminal_offset
        elif numb == 'CTR':
            numb = config.tit_mole._CTR + config.terminal_offset
        if isinstance(numb, str):
            try:
                numb = int(numb)
            except ValueError:
                raise Exception('Unknown site')
        return numb

    def getParameters(self):
        """Get the parameters used in the calculations
        """
        return config.tit_mole.getDelPhi()
    
    def __getitem__(self, numb):
        numb = self.correct_site_numb(numb)
        return self._pKas[numb]

    def __str__(self):
        output = '  Site   Name      pK'
        sites = config.tit_mole.getSites()

        for site in self._pKas:
            sitename = sites[site].getName()
            pk = self._pKas[site]
            if pk != '-':
                pk = round(pk, 2)
            else:
                pk = 'Not In Range'
            site = convertTermini(site)
            output += '\n{0:>6}    {1:3}    {2:>}'.format(site,
                                                          sitename, pk)
        return output


def CLI():
    # Assignment of global variables
    parameters, debug = checkParsedInput()
    Titration(parameters, sites=None, debug=debug)
    print('CLI exited successfully')


if __name__ == "__main__":
    CLI()



import config
from multiprocessing import Pool, Manager, Value
from time import time
from datetime import timedelta, datetime
from sys import stdout


def configRefresh(config, pb_time, njobs):
    config.pb_time = pb_time
    config.njobs = njobs


def startPoolProcesses(targetFunction, iterable_job_arguments_list,
                       ncpus, assign='distributed', merged_results=False):
    """Start a pool of ncpus processes that execute the targetFunction
    with the argument iterable_job_arguments_list

    Args:
      targetFunction (function): function to be execute in each process
      iterable_job_arguments_list (iterable): object enumerating
        the arguments to be passed to the targetFunction.
        when using assing='ordered', it can not be a generator
      ncpus  (int): number of processes to launch
      assign (str): mode to split the jobs among processes
        Example: iterable_job_arguments_list = [0, 1, 2, 3, 4], ncpus = 2
        'distributed' p0 = [0, 2, 4] p1 = [1, 3]
        'ordered' p0 = [0, 1, 2] p1 = [3, 4]

    Ensures:
      unpacked_results (list): list of lists containing the returned
        results from the executed jobs. the order is dependent
        only on $assign and $ncpus but not on the execution time.
    """
    jobs = [[] for i in range(ncpus)]
    results = []
    i = -1
    if assign == 'distributed':
        config.njobs = 0
        for tautomer in iterable_job_arguments_list:
            i += 1
            jobs[i % ncpus].append(i)
            config.njobs += 1
    elif assign == 'ordered':
        max_njobs = len(iterable_job_arguments_list) / ncpus
        ncores_extra_load = len(iterable_job_arguments_list) % ncpus
        core = 0
        core_jobs = 0
        extra_load = False
        for tautomer in iterable_job_arguments_list:
            i += 1
            if core_jobs == max_njobs:
                if ncores_extra_load > core and not extra_load:
                    core_jobs -= 1
                    extra_load = True
                else:
                    core_jobs = 0
                    core += 1
                    extra_load = False
            jobs[core].append(i)
            core_jobs += 1

    config.total_jobs = config.njobs
    pb_time = Manager().list()
    njobs = Value('i', config.njobs)
    pool = Pool(processes=ncpus, initializer=configRefresh,
                initargs=(config, pb_time, njobs))
    for job in jobs:
        result = pool.apply_async(targetFunction, args=(job, ))
        results.append(result)
        # Easier debug of the loop but fails afterwards
        #targetFunction(job)

    pool.close()
    pool.join()
    #print 'exit'
    #exit()
    
    unpacked_results = []
    for results_percore in results:
        result = results_percore.get()
        if merged_results:
            unpacked_results += result
        else:
            unpacked_results.append(result)
    return unpacked_results


def runDelPhiSims(job_list):
    """Run sets of simulations of tautomers included in job_list

    Args:
      job_list (list): list of tautomer numbers to be analysed.
        The use of numbers is needed since class objects
        can not be pickled and therefore can not be parsed
        in multiprocessing.

    Ensures:
      results (list)
        tauname (str):   the tautomer name
        sitenum (int):   site id number
        esolvM  (float): the Molecule solvation energy
        sitpotM (list):  the site potential in the Molecule
          the index corresponds to the atoms in the site
        esolvS  (float): the Site solvation energy
        sitpotS (list):  the site potential in the Site
          the index corresponds to the atoms in the site
    """
    if config.debug:
        # each process will run in its own directory
        #os.system('mkdir -p core_{0}'.format('_'.join([str(i) for i in job_list])))
        #os.chdir('core_{0}'.format('_'.join([str(i) for i in job_list])))

        # TODO: file name may be too long. needs fix
        pass


    results = []
    for tau_number in job_list:
        time1 = time()
        tauname, sitenum, esolvM, sitpotM, esolvS, sitpotS = calcPotential(tau_number)
        results.append([tauname, sitenum, esolvM, sitpotM, esolvS, sitpotS])
        time2 = time()

        config.njobs.value -= 1
        njobs = config.njobs.value
        config.pb_time.append(time2 - time1)
        left = sum(config.pb_time) / len(config.pb_time) * njobs / config.params['ncpus']
        if left > 3600:
            left_time = '~{0}h'.format(int(left / 3600.0))
        elif left > 60:
            left_time = '~{0}m'.format(int(left / 60.0))
        else:
            left_time = '{0}s'.format(int(left))
        
        end = datetime.now() + timedelta(seconds=left)
        end_time = end.strftime('%H:%M:%S %d/%m/%Y')
        stdout.write('\rPB Runs: {0:3} {1:<10} '
                     'Run {2:5} of {3:<10} '
                     'Ends in {4} at {5}'.format(tauname, sitenum,
                                                 config.total_jobs - njobs,
                                                 config.total_jobs,
                                                 left_time,
                                                 end_time))
        stdout.flush()

    stdout.write('\rPB Runs Ended{0:>70}'.format(''))
    stdout.flush()

    return results


def calcPotential(taut):
    """Run two DelPhi simulations: one for the tautomer and
    other for the same tautomer when all other sites are neutral.

    Args:
      taut (int): number of the tautomer

    Ensures:
      tauname (str):   the tautomer name
      sitenum (int):   site id number
      esolvM  (float): the Molecule solvation energy
      sitpotM (list):  the site potential in the Molecule
        the index corresponds to the atoms in the site
      esolvS  (float): the Site solvation energy
      sitpotS (list):  the site potential in the Site
        the index corresponds to the atoms in the site
    """
    # get the tautomer object from the tautomer number
    # transformation needed due to multiprocessing
    taut = config.tit_mole.getTautomerNumber(taut)

    # Whole Molecule with zero charge except for the tautomer being evaluated
    e_solvationM, sitpotM = taut.CalcPotentialTitratingMolecule()

    if config.debug:
        print(('finished Whole Molecule', taut._name, e_solvationM))

    # Single Site Tautomer
    e_solvationS, sitpotS = taut.CalcPotentialTautomer()

    if config.debug:
        print(('finished Tautomer', taut._name, e_solvationS))

    return (taut._name, taut.getSiteResNumber(), e_solvationM,
            sitpotM, e_solvationS, sitpotS)


def runInteractionCalcs(job_list):
    """Run interaction calculations between
    sites in job_list

    Args:
      job_list (list)
        interaction_number (int): interaction number to be analysed.
          The use of numbers is needed since class objects
          can not be pickled and therefore can not be parsed
          in multiprocessing.

    Ensures:
      results (list)
        interaction_energies (str): .dat formatted energy
          interactions between all tautomers of two sites
    """
    results = []
    for interaction_number in job_list:
        interaction_energies = config.tit_mole.calcInteractionNumber(interaction_number)
        results += interaction_energies

    return results


def runMCCalcs(job_list):
    """
    """
    results = []
        
    for pH_index in job_list:
        mc_output = config.tit_mole.parallelMCrun(pH_index)
        results.append(mc_output) 

    return results



import pytest
import os
import subprocess as sb
import sys
sys.path.insert(1, '../')

ncpus = 1

# run coverage.sh to generate coverage

def runTest(path, ncpus, results):
    os.system("rm -f {0}/*out {0}/clean*pqr {0}/TMP.gro {0}/delphi_in_*pdb".format(path))
    results_lines = results.split('\n')[1:-1]
    sb.Popen("cd {0}; "
             "sed -i 's/ncpus .*/ncpus           = {1}/' parameters.dat; "
             "python3 -m coverage erase; "
             "python3 -m coverage run ../../../pypka.py parameters.dat".format(path, ncpus), shell=True).wait()
    checkOutput('{0}/pKas.out'.format(path), results_lines)    

def checkOutput(filename, results_lines):
    with open(filename) as f:
        c = -1
        for line in f:
            c += 1
            line = line.strip()
            assert line == results_lines[c]
    assert c + 1 == len(results_lines)

def checkAPIResult(pKa, results):
    i = -1
    for site in pKa:
        result = '{0} {1} {2}'.format(site, pKa[site], pKa.getProtState(site, 7))
        i += 1
        assert result == results[i]
    assert i + 1 == len(results)

class TestCLI(object):
    def test_cli_ktp_gro(self):
        path = "ktp/ktp_gro"
        results = """
1 NTR 7.900646106133701
1 TYR 9.88522556057786
2 CTR 3.123850715655238
        """
        runTest(path, ncpus, results)

    def test_cli_ktp_pdb_allsites(self):
        path = "ktp/ktp_pdb_allsites"
        results = """
1 NTR 7.9037275985663085
1 TYR 9.783097263247802
2 CTR 3.3040036750415207
        """
        runTest(path, ncpus, results)

    def test_cli_ktp_pdb_allsites_noclean(self):
        path = "ktp/ktp_pdb_allsites_noclean"
        results = """
1 NTR 7.900646106133701
1 TYR 9.88522556057786
2 CTR 3.123850715655238
        """
        runTest(path, ncpus, results)

    def test_cli_ktp_pdb_onlytermini(self):
        path = 'ktp/ktp_pdb_onlytermini'
        results = """
1 NTR 7.9001039320976965
2 CTR 3.2990234444483972
        """
        runTest(path, ncpus, results)

    def test_cli_ktp_pdb_onlytermini_noclean(self):
        path = 'ktp/ktp_pdb_onlytermini_noclean'
        results = """
1 NTR 8.167718884417926
2 CTR 3.390519247038917
        """
        runTest(path, ncpus, results)

    def test_cli_lyso_gro(self):
        path = "lyso/lyso_gro"
        results = """
1 NTR 7.63661126098686
18 ASP 3.141601921175308
35 GLU 4.663011344104504
48 ASP 2.9639603215645716
66 ASP 3.2554706185808753
129 CTR 2.353209368853925
        """
        runTest(path, ncpus, results)

    def test_cli_lyso_pdb_sites(self):
        path = "lyso/lyso_pdb_sites"
        results = """
1 NTR 7.505477696120323
18 ASP 3.1651915544951166
35 GLU 4.610836750719833
48 ASP 2.258441855151931
66 ASP 1.871335236896869
129 CTR 2.3040367135982893
        """
        runTest(path, ncpus, results)

    def test_cli_lyso_pdb_sites_noclean(self):
        path = "lyso/lyso_pdb_sites_noclean"
        results = """
1 NTR 7.39335708373817
18 ASP 2.7816106018108915
35 GLU 4.271223894654857
48 ASP 2.6544965096521933
66 ASP 3.0112861335754557
129 CTR 1.9409245098602885
        """
        runTest(path, ncpus, results)

    def test_cli_lyso_pdb_all(self):
        path = "lyso/lyso_pdb_all"
        results = """
1 NTR 7.384144013092099
1 LYS 10.391515199306076
7 GLU 2.880921222215613
13 LYS 100.0
15 HIS 5.861356475494928
18 ASP 2.6945728178749397
20 TYR 9.910664161785833
23 TYR 9.392332137365454
24 SER 100.0
33 LYS 10.510393195704687
35 GLU 4.179904580152671
36 SER 100.0
40 THR 100.0
43 THR 100.0
47 THR 100.0
48 ASP 1.8672505767012688
50 SER 100.0
51 THR 100.0
52 ASP 2.1472304416360166
53 TYR 10.743180477033484
60 SER 100.0
66 ASP 2.3464474331287817
69 THR 100.0
72 SER 100.0
81 SER 100.0
85 SER 100.0
86 SER 100.0
87 ASP 2.0476639876603344
89 THR 100.0
91 SER 100.0
96 LYS 100.0
97 LYS 100.0
100 SER 100.0
101 ASP 3.31081873557565
116 LYS 10.255162473280647
118 THR 100.0
119 ASP 2.3955112075754372
129 CTR 1.7035346555851842
        """
        runTest(path, ncpus, results)

    def test_cli_lyso_pdb_all_noclean(self):
        path = "lyso/lyso_pdb_all_noclean"
        results = """
1 NTR 7.384144013092099
1 LYS 10.391515199306076
7 GLU 2.880921222215613
13 LYS 100.0
15 HIS 5.861356475494928
18 ASP 2.6945728178749397
20 TYR 9.910664161785833
23 TYR 9.392332137365454
24 SER 100.0
33 LYS 10.510393195704687
35 GLU 4.179904580152671
36 SER 100.0
40 THR 100.0
43 THR 100.0
47 THR 100.0
48 ASP 1.8672505767012688
50 SER 100.0
51 THR 100.0
52 ASP 2.1472304416360166
53 TYR 10.743180477033484
60 SER 100.0
66 ASP 2.3464474331287817
69 THR 100.0
72 SER 100.0
81 SER 100.0
85 SER 100.0
86 SER 100.0
87 ASP 2.0476639876603344
89 THR 100.0
91 SER 100.0
96 LYS 100.0
97 LYS 100.0
100 SER 100.0
101 ASP 3.31081873557565
116 LYS 10.255162473280647
118 THR 100.0
119 ASP 2.3955112075754372
129 CTR 1.7035346555851842
        """
        runTest(path, ncpus, results)

    def test_cli_pHLIP_gro(self):
        path = "pHLIP/pHLIP_gro"
        results = """
769 NTR 10.767904983032684
770 CYS 100.0
771 GLU 3.997527399091152
782 ASP 100.0
793 ASP 100.0
799 ASP 1.9413177479539034
801 ASP 0.9902989821882952
802 GLU 4.059942648076733
804 CTR 3.8417877441756705
        """
        runTest(path, ncpus, results)

    def test_cli_pHLIP_pdb_all(self):
        path = "pHLIP/pHLIP_pdb_all"
        results = """
769 NTR 11.202571597654359
770 CYS 24.21238816620012
771 GLU 4.010786085385137
776 TYR 18.77569436939371
780 TYR 24.348420389190892
782 ASP 100.0
786 THR 100.0
787 THR 100.0
793 ASP 22.658592170556883
799 ASP 0.37064577017674316
801 ASP 1.4856797431942301
802 GLU 3.965613455945925
804 CTR 4.289631904096492
        """
        runTest(path, ncpus, results)


class TestAPI(object):
    def test_api_ktp_gro(self):
        from ..pypka import Titration
        os.system('rm -f LOG* *out *gro *pdb *pqr *crg *sites cent contributions interactions.dat pkint')
        results = """
NTR 7.900646106133701 ('undefined', 0.8883318942980771)
1 9.88522556057786 ('protonated', 0.9986992041484555)
CTR 3.123850715655238 ('deprotonated', 0.00013298203005695505)
        """
        results = results.split('\n')[1:-1]
        parameters = {
            'structure'     : 'ktp/ktp_gro/ktp.gro',     # MANDATORY
            'epsin'         : 2,
            'ionicstr'      : 0.1,
            'pbc_dimensions': 0,
            'temp'          : 310,
            'grid_fill'     : 0.8,         # FUTURE VERSION
            'ncpus'         : ncpus,
            'pH'            : '-5,15',
            'pHstep'        : 0.2,
            'logfile'       : 'LOGFILE',
            'scaleM'        : 4,
            'scaleP'        : 1,
            'gsize'         : 81,
            'convergence'   : 0.01,
            'nlit'          : 500,
            'cutoff'        : -1,
            'relfac'        : 0.0,            
            'output'        : 'pKas.out'
        }
        sites = {'A': ('1N', '1', '2C')}
        pKa = Titration(parameters, sites=sites)
        checkAPIResult(pKa, results)

    def test_api_ktp_pdb_allsites_clean(self):
        from ..pypka import Titration
        os.system('rm -f *out *gro *pdb *pqr *crg *sites cent contributions interactions.dat pkint')
        results = """
NTR 7.904092836420442 ('undefined', 0.8891167479866279)
1 9.780713528788414 ('protonated', 0.9983458781425558)
CTR 3.3033136197599586 ('deprotonated', 0.00020101400276712804)
        """
        results = results.split('\n')[1:-1]
        parameters = {
            'structure'     : 'ktp/ktp_pdb_allsites/ktp.pdb',
            'clean_pdb'     : 'yes',
            'epsin'         : 2,
            'ionicstr'      : 0.1,
            'pbc_dimensions': 0,
            'temp'          : 310,
            'grid_fill'     : 0.8,
            'ncpus'         : ncpus,
            'pH'            : '0,15',
            'pHstep'        : 0.2,
            'logfile'       : 'LOGFILE',
            'scaleM'        : 4,
            'scaleP'        : 1,
            'gsize'        : 81,
            'convergence'   : 0.01,
            'nlit'          : 300,
            'cutoff'        : -1,
            'relfac'        : 0.0,
            'output'        : 'pKas.out'
        }
        sites = 'all'           
        pKa = Titration(parameters, sites=sites)
        checkAPIResult(pKa, results)

    def test_api_ktp_pdb_allsites_noclean(self):
        from ..pypka import Titration
        os.system('rm -rf LOG* *out *gro *pdb *pqr *crg *sites '
                  'cent contributions interactions.dat pkint LOG* __pycache__ *pyc')
        results = """
NTR 7.900646106133701 ('undefined', 0.8883318942980771)
1 9.88522556057786 ('protonated', 0.9986992041484555)
CTR 3.123850715655238 ('deprotonated', 0.00013298203005695505)
        """
        results = results.split('\n')[1:-1]
        parameters = {
            'structure'     : 'ktp/ktp_pdb_allsites_noclean/ktp_noclean.pdb',     # MANDATORY
            'clean_pdb'     : 'no',
            'epsin'         : 2,
            'ionicstr'      : 0.1,
            'pbc_dimensions': 0,
            'temp'          : 310,
            'grid_fill'     : 0.8,         # FUTURE VERSION
            'ncpus'         : ncpus,
            'pH'            : '-5,15',
            'pHstep'        : 0.2,
            'logfile'       : 'LOGFILE',
            'scaleM'        : 4,
            'scaleP'        : 1,
            'gsize'         : 81,
            'convergence'   : 0.01,
            'nlit'          : 500,
            'cutoff'        : -1,
            'relfac'        : 0.0,            
            'output'        : 'pKas.out'
        }
        sites = 'all'           
        pKa = Titration(parameters, sites=sites)
        checkAPIResult(pKa, results)

    def test_api_ktp_pdb_onlytermini(self):
        from ..pypka import Titration
        os.system('rm -rf LOG* *out *gro *pdb *pqr *crg *sites cent '
                  'contributions interactions.dat pkint LOG* __pycache__ *pyc')
        results = """
NTR 7.9001039320976965 ('undefined', 0.8882079948571108)
CTR 3.2990234444483972 ('deprotonated', 0.0001990384562145051)
        """
        results = results.split('\n')[1:-1]
        parameters = {
            'structure'     : 'ktp/ktp_pdb_onlytermini/ktp.pdb',     # MANDATORY            
            'clean_pdb'     : 'yes',
            'epsin'         : 2,
            'ionicstr'      : 0.1,
            'pbc_dimensions': 0,
            'temp'          : 310,
            'grid_fill'     : 0.8,         # FUTURE VERSION
            'ncpus'         : ncpus,
            'pH'            : '0,15',
            'pHstep'        : 0.25,
            'logfile'       : 'LOGFILE',
            'scaleM'        : 4,
            'scaleP'        : 1,
            'gsize'         : 81,
            'convergence'   : 0.01,
            'nlit'          : 300,
            'cutoff'        : -1,
            'relfac'        : 0.0,            
            'output'        : 'pKas.out'
        }
        sites = {'A': ('1N', '2C')}
        pKa = Titration(parameters, sites=sites)
        checkAPIResult(pKa, results)

    def test_api_ktp_pdb_onlytermini_noclean(self):
        from ..pypka import Titration
        os.system('rm -rf LOG* *out *gro *pdb *pqr *crg *sites cent '
                  'contributions interactions.dat pkint LOG* __pycache__ *pyc')
        results = """
NTR 8.167718884417926 ('protonated', 0.9363608510357995)
CTR 3.390519247038917 ('deprotonated', 0.0002457041692844058)
        """
        results = results.split('\n')[1:-1]
        parameters = {
            'structure'     : 'ktp/ktp_pdb_onlytermini_noclean/ktp_noclean.pdb',     # MANDATORY            
            'clean_pdb'     : 'no',
            'epsin'         : 2,
            'ionicstr'      : 0.1,
            'pbc_dimensions': 0,
            'temp'          : 310,
            'grid_fill'     : 0.8,         # FUTURE VERSION
            'ncpus'         : ncpus,
            'pH'            : '0,15',
            'pHstep'        : 0.25,
            'logfile'       : 'LOGFILE',
            'scaleM'        : 4,
            'scaleP'        : 1,
            'gsize'         : 81,
            'convergence'   : 0.01,
            'nlit'          : 300,
            'cutoff'        : -1,
            'relfac'        : 0.0,
            'output'        : 'pKas.out'
        }
        sites = {'A': ('1N', '2C')}
        pKa = Titration(parameters, sites=sites)

        checkAPIResult(pKa, results)



import argparse
import os
import config
import log


def drange(dmin, dmax, step):
    """Decimal Range

    Requires:
      dmin (float or int) is the range minimum value
      dmax (float or int) is the range maximum value
      step (float or int) is the range step value

    Ensures:
      lrange (list)

    Example:
      drange(5, 7.5, 0.5)
        -> [5.0, 5.5, 6.0, 6.5, 7.0, 7.5]

    """
    lrange = []
    dmin, dmax = float(dmin), float(dmax)
    step = float(step)
    inv_step = step ** -1
    dmin_int = int(int(dmin) * inv_step)
    dmax_int = int(int(dmax) * inv_step) + 1
    for i in range(dmin_int, dmax_int):
        lrange.append(i / inv_step)
    return lrange


def setParameter(param, value):
    if param in list(config.input_conversion.keys()):
        param = config.input_conversion[param]
    config.params[param] = value


def getParameter(param):
    if param in list(config.input_conversion.keys()):
        param = config.input_conversion[param]
    return config.params[param]


def inputParametersFilter(settings):
    """
    Check if input parameters are valid

    Mandatory Parameters:
      - Dielectric Constant
      - Ionic Strength

    """
    config.pid = os.getpid()
    config.script_dir = os.path.dirname(__file__)
    param_names = list(settings.keys())
    # MANDATORY #
    mandatory_params = ('structure', 'epsin', 'ionicstr',
                        'pbc_dimensions', 'ncpus', 'sites_A')
    for param in mandatory_params:
        if param not in param_names:
            log.requiredParameterError(param)
        if type(settings[param]) in (float, int):
            settings[param] = str(settings[param])
        if len(settings[param]) == 0:
            log.requiredParameterError(param)

    # Check parameter conditions: parameter is integer
    integer_params = ('gsize', 'seed', 'ncpus')
    for param in integer_params:
        if param in param_names:
            try:
                param_value = int(settings[param])
            except ValueError:
                log.inputVariableError(param,
                                       'an integer.', '')
            setParameter(param, param_value)

    # Check parameter conditions: parameter is float
    float_params = ('scaleP', 'scaleM', 'ionicstr', 'convergence', 'nlit',
                    'nonit', 'relfac', 'relpar', 'pHstep', 'epssol',
                    'temp', 'epsin', 'slice', 'cutoff')
    for param in float_params:
        if param in param_names:
            try:
                param_value = float(settings[param])
            except ValueError:
                log.inputVariableError(param,
                                       'a float.', '')
            setParameter(param, param_value)

    # Check parameter conditions: parameter > 0
    # These parameters have already been checked for type int or float
    great_params = ('scaleP', 'scaleM', 'convergence', 'pHstep', 'gsize',
                    'ncpus', 'temp', 'grid_fill', 'pH_step')
    for param in great_params:
        if param in param_names:
            param_value = getParameter(param)
            if not param_value > 0:                    
                log.inputVariableError(param,
                                       'greater than zero.', '')
            setParameter(param, param_value)

    # Check parameters conditions: parameter is boolean
    bool_params = ('pbx', 'pby', 'clean_pdb')
    for param in bool_params:
        if param in param_names:
            param_value = settings[param]
            if param_value == 'yes':
                param_value = True
            elif param_value == 'no':
                param_value = False
            else:
                log.inputVariableError(param,
                                       'either "yes" or "no".', '')

            setParameter(param, param_value)

    # Check particular parameter conditions
    if 'bndcon' in param_names:
        if str(settings['bndcon']) not in ('1', '2', '3', '4'):

            log.inputVariableError('bndcon',
                                   '1 (zero), 2(dipolar),'
                                   ' 3(focusing) or 4 (coulombic).', '')
        
        else:
            setParameter('bndcon', int(settings['bndcon']))

    if 'precision' in param_names:
       if settings['precision'] not in ('single', 'double'):

           log.inputVariableError('precision',
                                  'either "single" or "double".', '')
       else:
           setParameter('precision', settings['precision'])

    if 'ffID' in param_names:
        setParameter('ffID', settings['ffID'])
        if settings['ffID'] not in ('G54A7'):  # for now only GROMOS FF
            log.inputVariableError('ffID',
                                   'equal to "G54A7".', '')
    
    if 'ffinput' in param_names:
        setParameter('ffinput', settings['ffinput'])
        if settings['ffinput'] not in ('GROMOS', 'AMBER', 'CHARMM'):
            log.inputVariableError('ffinput',
                                   'either "GROMOS", "AMBER" or "CHARMM".', '')
    
    file_path = os.path.join(config.script_dir, config.params['ffID'])
    config.f_crg = '{0}/DataBaseT.crg'.format(file_path)
    config.f_siz = '{0}/DataBaseT.siz'.format(file_path)

    if settings['pbc_dimensions'] not in ('0', '2'):
        log.inputVariableError('pbc_dimensions',
                               'either "0" or "2".', '')
    else:
        param_value = int(settings['pbc_dimensions'])
        setParameter('pbc_dimensions', param_value)

    if 'nanoshaper' in settings:
        if settings['nanoshaper'] in ('0', '1', '2', '3', '4', '-1'):
            log.inputVariableError('nanoshaper',
                                   'either "-1" to turn of nanoshaper or "0" to turn on nanoshaper.', '')
        else:
            param_value = int(settings['nanoshaper'])
            config.nanoshaper = param_value
        
    # Needs to accept both a single value and a range
    if 'pH' in param_names:
        pH_parts = settings['pH'].split(',')
        if len(pH_parts) > 1:
            try:
                pHmin = float(pH_parts[0])
                pHmax = float(pH_parts[1])
                setParameter('pHmin', pHmin)
                setParameter('pHmax', pHmax)
            except ValueError:
                log.inputVariableError('pH',
                                       'a float.', '')
        else:
            try:
                pHmin = float(pH_parts[0])
                pHmax = float(pH_parts[0])
                setParameter('pHmin', pHmin)
                setParameter('pHmax', pHmax)
            except ValueError:
                log.inputVariableError('pH',
                                       'a float.', '')
            setParameter('pH', [param_value])
        if pHmin >= pHmax:
            log.inputVariableError('pHmax',
                                   'a float greater than pHmin.', '')

    # Declare IO Files
    # Input .pdb File
    config.f_in = settings['structure']
    f_in_parts = settings['structure'].split('.')
    if len(f_in_parts) <= 1:
        log.inputVariableError('structure',
                               'a string containing a file extension.',
                               'Ex: structure.pdb or structure.gro')

    extension = f_in_parts[1].lower().replace('pqr', 'pdb')
    if extension not in ('gro', 'pdb'):
        log.inputVariableError('structure',
                               'a string containing a valid file extension.',
                               'Ex: structure.pdb or structure.gro or structure.pqr')

    config.f_in_extension = extension

    # Output pKs File
    if 'output' in param_names:
        config.f_out = settings['output']
    else:
        outputname = settings['structure'].split('.')[0]
        config.f_out = outputname

    # Output Titration File
    if 'titration_output' in param_names:
        config.f_prot_out = settings['titration_output']

    # Output log File
    if 'logfile' in param_names:
        config.f_log = settings['logfile']  # default: "LOG"

    # Check coherence between variables
    if getParameter('pbc_dim') == 2:
        if getParameter('relfac') != 0.2 and 'relfac' not in settings:
            setParameter('relfac', 0.2)
        if getParameter('nonit') != 5 and 'nonit' not in settings:
            setParameter('nonit', 5)

    if 'lipid_definition' in settings:
        for i in settings['lipid_definition']:
            resname = settings['lipid_definition'][i]
            config.lipids[i] = resname
            if resname in config.lipid_residues:
                resname_i = config.lipid_residues.index(resname)
                del config.lipid_residues[resname_i]
    return


def readSettings(filename):
    """Reads the settings file.

    This file should have the following format:
     - commented lines should being with a '#'
     - every parameter should be declared as such: name = value

    All parameter values are interpreted as strings, however, a type
    check is later performed for each declared input value.

    All parameter names not recognizable are reported as a warning.
    """
    parameters = {}
    parameters['lipid_definition'] = {}
    with open(filename) as f:
        nline = 0
        for line in f:
            nline += 1
            if len(line.strip()) > 0 and line[0] != '#':
                parts = line.split('=')
                param_name = parts[0].strip()
                param_value = '='.join(parts[1:]).strip()
                if 'lipid_definition' in param_name:
                    parts = param_value.split(':')
                    old_name = parts[0]
                    new_name = parts[1]
                    parameters['lipid_definition'][old_name] = new_name
                elif (len(parts) != 2 or
                      not len(param_name) > 0 or
                      not len(param_value) > 0):
                    raise IOError('Incorrect format in line {0} of file {1}: '
                                  '\n{1}#{0}: {2}'.format(nline, filename, line))
                else:
                    parameters[param_name] = param_value

    # Search for all titrable sites in different chains
    sites = {}
    for param_name in parameters:
        if 'site' in param_name:
            chain = param_name.split('_')[1]
            chain_sites = parameters[param_name].split(', ')
            if chain_sites != ['all']:
                sites[chain] = chain_sites
    config.sites = sites

    return parameters


def checkParsedInput():
    """Gets the CLI arguments and interprets them"""

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter, description="""
Object-Oriented Script to Calculate the pKint of each site \
as well as the pairwise energies
Requires:
DelPhi2Py module installation
Nanoshaper and cppSolver are optional libraries

Objects:
DelPhiParams stores the DelPhi input parameters like a .prm file

TitratingMolecule is the molecule which has more than one Site

Site

Tautomer

Example:
python pypka.py test.pdb test.dat -o pKas.out --debug

    """)

    # Mandatory Arguments
    parser.add_argument('settings', help=' settings file name',
                        default="settings.dat", action='store')

    # Optional Arguments
    parser.add_argument('--debug', help='activation of the debug mode '
                        'to print extra information', action='store_true')

    args = parser.parse_args()

    # Apply some criteria to input arguments
    if not os.path.isfile(args.settings):
        raise IOError('File {0} does not exist.'.format(args.settings))

    # Read Settings File
    config.f_dat = args.settings
    parameters = readSettings(args.settings)

    return parameters, args.debug



import os
import log
from formats import (read_pdb_line, read_pqr_line, read_gro_line,
                     correct_names, new_pqr_line, new_gro_line, pdb2gro)
import config
from copy import copy

def inputPDBCheck(filename, sites):
    """
    chains_length, chains_res
    """
    if filename[-3:] in ('pdb', 'pqr'):
        filetype = 'pdb'
    elif filename[-3:] == 'gro':
        filetype = 'gro'
    else:
        raise Exception('Input file must be either a pdb or a gro.')

    chains_length = {}
    chains_res = {}
    done = {}

    for site in sites['A']:
        if site[-1] == 'C':
            resnumb = site[:-1]
            done[resnumb] = 'CTR'
        if site[-1] == 'N':
            resnumb = site[:-1]
            done[resnumb] = 'NTR'

    if filetype == 'pdb' and not config.params['clean_pdb']:
        new_gro_header = 'CREATED within PyPka\n'
        new_gro_body = ''
    with open(filename) as f:
        last_chain = ''
        chain_length = 0

        nline = 0
        maxnlines = 0
        atom_number = 0
        for line in f:
            nline += 1
            atom_line = False
            if filetype == 'pdb':
                if 'ATOM ' == line[0:5]:
                    atom_line = True
                    chain_length += 1
                    (aname, anumb, resname,
                     chain, resnumb, x, y, z) = read_pdb_line(line)
                    atom_number += 1
                    if not config.params['clean_pdb']:
                        if len(aname) > 2 and \
                           aname[1] == 'H' and \
                           aname[0] in ('1', '2'):
                            aname = aname[1:] + aname[0]
                        new_gro_body += new_gro_line(anumb, aname,
                                                     resname, resnumb,
                                                     x / 10.0, y / 10, z / 10)
                elif 'CRYST1' in line:
                    tmp = line.split()[1:4]
                    box = (float(tmp[0]), float(tmp[1]), float(tmp[2]))
                    new_gro_footer = '{0:10.5f}{1:10.5f}{2:10.5f}\n'.format(box[0] / 10.0,
                                                                            box[1] / 10.0,
                                                                            box[2] / 10.0)

            elif filetype == 'gro':
                if nline > 2 and nline < maxnlines:
                    (aname, anumb, resname, resnumb, x, y, z) = read_gro_line(line)
                    chain = 'A'
                    atom_line = True
                elif nline == 2:
                    natoms = int(line.strip())
                    maxnlines = natoms + 3

            if atom_line:
                if chain == ' ':
                    chain = 'A'

                if chain_length == 1:
                    last_chain = chain

                if chain != last_chain and chain_length != 1:
                    chains_length[last_chain] = chain_length
                    chains_res[chain] = done
                    chain_length = 0
                    last_chain = chain

                if chain in sites and \
                   resnumb not in done and \
                   str(resnumb) in sites[chain]:
                    done[resnumb] = resname

    if filetype == 'pdb' and not config.params['clean_pdb']:
        new_gro_header += '{0}\n'.format(atom_number)
        with open('TMP.gro', 'w') as f:
            f.write(new_gro_header + new_gro_body + new_gro_footer)

    chains_length[last_chain] = chain_length
    chains_res[chain] = done

    return chains_length, chains_res


def cleanPDB(pdb_filename, pdb2pqr_path, chains_res,
             inputpqr, outputpqr, sites):
    """
    """

    inputpdbfile = removeMembrane(pdb_filename)

    logfile = 'LOG_pdb2pqr'
    errfile = 'LOG_pdb2pqr_err'

    log.redirectOutput("start", logfile)
    log.redirectErr("start", errfile)

    sites_numbs = list(sites.keys())

    # CTR O1/O2 will be deleted and a O/OXT will be added
    os.system('python2 {0} {1} {2} --ff {3} --ffout GROMOS '
              '--drop-water -v --chain'.format(pdb2pqr_path,
                                               inputpdbfile, inputpqr,
                                               config.params['ffinput']))

    log.redirectOutput("stop", logfile)
    log.redirectErr("stop", errfile)

    CYS_bridges = []
    with open('LOG_pdb2pqr') as f:
        for line in f:
            if 'patched with CYX' in line:
                parts = line.split('patched')[0].replace('PATCH INFO: ', '').split()
                CYS_bridges.append(int(parts[-1]))

    new_pdb_text = ''
    aposition = 0
    #ntr_trigger = True
    with open(inputpqr) as f:
        for line in f:
            if "ATOM" in line[:4]:
                (aname, anumb, resname, chain, resnumb, x, y,
                 z, charge, radius) = read_pqr_line(line)
                aposition += 1

                aname, resname = correct_names(sites_numbs, resnumb,
                                               resname, aname, sites_numbs)

                new_pdb_text += new_pqr_line(aposition, aname, resname,
                                             resnumb, x, y, z, charge, radius)

    #            if ntr_trigger and :
    #                config.tit_mole._NTR = resnumb
    
    chains = copy(chains_res['A'])
    #for res in chains:
    #    if str(res) > config.terminal_offset:
    #        print((res, chains_res['A'][res]))
    #        del chains_res['A'][res]
    #chains_res['A'][config.tit_mole._NTR] = 'NTR'
    #chains_res['A'][config.tit_mole._CTR] = 'CTR'

    with open('cleaned.pqr', 'w') as f_new:
        f_new.write(new_pdb_text)

    

    # TODO: .sites files should not exist
    # adapt addHtaut to work without these
    sites_addHtaut = ''
    for res in chains_res['A']:
        if chains_res['A'][res] != 'NTR' and res not in CYS_bridges:
            sites_addHtaut += '{0}_{1},'.format(res, chains_res['A'][res])

    if len(sites_addHtaut) > 0 and sites_addHtaut[-1] == ',':
        sites_addHtaut = sites_addHtaut[:-1]

    if config.debug:
        logfile = 'LOG_addHtaut'
        print(sites_addHtaut)
    else:
        logfile = config.f_log
    log.redirectErr("start", logfile)

    os.system('{0}/addHtaut cleaned.pqr {1} > '
              'cleaned_tau.pqr'.format(config.script_dir, sites_addHtaut))

    log.redirectErr("stop", logfile)
    log.checkDelPhiErrors(logfile, 'addHtaut')

    with open(pdb_filename) as f:
        for line in f:
            if 'CRYST1' in line:
                tmp = line.split()[1:4]
                box = (float(tmp[0]), float(tmp[1]), float(tmp[2]))

    pdb2gro(outputpqr, "TMP.gro", box, sites, pqr=True)

    if config.params['pbc_dim'] == 2:
        addMembrane("TMP.gro", pdb_filename)

    tmpfiles = ('LOG_pdb2pqr', 'LOG_pdb2pqr_err', 'clean.pqr', 'cleaned.pqr',
                'cleaned_tau.pqr', 'input_clean.pdb', config.f_log)
    for filename in tmpfiles:
        if not config.debug and os.path.isfile(filename):
            os.remove(filename)


def removeMembrane(pdbfile):
    nomembrane_text = ''
    with open(pdbfile) as f:
        for line in f:
            if 'ATOM ' == line[0:5]:
                (aname, anumb, resname,
                 chain, resnumb, x, y, z) = read_pdb_line(line)
                if resname not in config.lipid_residues and chain in (' ', 'A'):
                    nomembrane_text += line
            else:
                nomembrane_text += line
    with open('tmp.tmp', 'w') as f_new:
        f_new.write(nomembrane_text)
    os.rename('tmp.tmp', 'input_clean.pdb')
    return 'input_clean.pdb'


def addMembrane(grofile, pdbfile):
    atom_number = 0
    new_file_header = ''
    new_file_body = ''
    new_file_footer = ''
    # Read the grofile with only the protein
    with open(grofile) as f:
        nline = 0
        maxnlines = 0
        for line in f:
            nline += 1
            if nline > 2 and nline < maxnlines:
                atom_number += 1
                (aname, anumb, resname, resnumb, x, y, z) = read_gro_line(line)
                new_file_body += new_gro_line(atom_number, aname, resname,
                                              resnumb, x, y, z)
            elif nline == 2:
                natoms = int(line.strip())
                maxnlines = natoms + 3
            elif nline == 1:
                new_file_header += line
            else:
                new_file_footer += line

    # Read the original pdb with the membrane
    with open(pdbfile) as f:
        for line in f:
            if 'ATOM ' == line[0:5]:
                (aname, anumb, resname,
                 chain, resnumb, x, y, z) = read_pdb_line(line)
                if resname in config.lipid_residues:
                    atom_number += 1
                    x, y, z = x/10, y/10, z/10
                    new_file_body += new_gro_line(atom_number, aname, resname,
                                                  resnumb, x, y, z)

                if resname in list(config.lipids.values()):
                    aname, resname, to_include = convert_FF_atomnames(aname, resname)
                    if to_include:
                        atom_number += 1
                        x, y, z = x/10, y/10, z/10
                        new_file_body += new_gro_line(atom_number, aname, resname,
                                                      resnumb, x, y, z)

    with open('tmp.tmp', 'w') as f_new:
        new_file_header += str(atom_number) + '\n'
        f_new.write(new_file_header + new_file_body + new_file_footer)
    os.rename('tmp.tmp', 'TMP.gro')


def convert_FF_atomnames(aname, resname):
    popc_resname = config.lipids['POPC']
    chol_resname = config.lipids['cholesterol']

    lookup = [((' N  ', popc_resname), ('NTM', 'CHL')),
              ((' C12', popc_resname), ('CB ', 'CHL')),
              ((' C13', popc_resname), ('CN1', 'CHL')),
              ((' C14', popc_resname), ('CN2', 'CHL')),
              ((' C15', popc_resname), ('CN3', 'CHL')),
              ((' C11', popc_resname), ('CA ', 'PJ2')),
              ((' P  ', popc_resname), ('P  ', 'PJ2')),
              ((' O13', popc_resname), ('OG ', 'PJ2')),
              ((' O14', popc_resname), ('OB ', 'PJ2')),
              ((' O12', popc_resname), ('OA ', 'PJ2')),
              ((' O11', popc_resname), ('OD ', 'PJ2')),
              ((' C1 ', popc_resname), ('CD ', 'POX')),
              ((' C2 ', popc_resname), ('CE ', 'POX')),
              ((' O21', popc_resname), ('OE ', 'POX')),
              ((' C21', popc_resname), ('C1A', 'POX')),
              ((' O22', popc_resname), ('O1A', 'POX')),
              ((' C22', popc_resname), ('C1B', 'POX')),
              ((' C3 ', popc_resname), ('CZ ', 'POX')),
              ((' O31', popc_resname), ('OZ ', 'POX')),
              ((' C31', popc_resname), ('C2A', 'POX')),
              ((' O32', popc_resname), ('O2A', 'POX')),
              ((' C32', popc_resname), ('C2B', 'POX')),
              ((' C23', popc_resname), ('C1C', 'POX')),
              ((' C24', popc_resname), ('C1D', 'POX')),
              ((' C25', popc_resname), ('C1E', 'POX')),
              ((' C26', popc_resname), ('C1F', 'POX')),
              ((' C27', popc_resname), ('C1G', 'POX')),
              ((' C28', popc_resname), ('C1H', 'POX')),
              ((' C29', popc_resname), ('C1I', 'POX')),
              (('0C21', popc_resname), ('C1J', 'POX')),
              (('1C21', popc_resname), ('C1K', 'POX')),
              (('2C21', popc_resname), ('C1L', 'POX')),
              (('3C21', popc_resname), ('C1M', 'POX')),
              (('4C21', popc_resname), ('C1N', 'POX')),
              (('5C21', popc_resname), ('C1O', 'POX')),
              (('6C21', popc_resname), ('C1P', 'POX')),
              (('7C21', popc_resname), ('C1Q', 'POX')),
              (('8C21', popc_resname), ('C1R', 'POX')),
              ((' C33', popc_resname), ('C2C', 'POX')),
              ((' C34', popc_resname), ('C2D', 'POX')),
              ((' C35', popc_resname), ('C2E', 'POX')),
              ((' C36', popc_resname), ('C2F', 'POX')),
              ((' C37', popc_resname), ('C2G', 'POX')),
              ((' C38', popc_resname), ('C2H', 'POX')),
              ((' C39', popc_resname), ('C2I', 'POX')),
              (('0C31', popc_resname), ('C2J', 'POX')),
              (('1C31', popc_resname), ('C2K', 'POX')),
              (('2C31', popc_resname), ('C2L', 'POX')),
              (('3C31', popc_resname), ('C2M', 'POX')),
              (('4C31', popc_resname), ('C2N', 'POX')),
              (('5C31', popc_resname), ('C2O', 'POX')),
              (('6C31', popc_resname), ('C2P', 'POX')),
              ((" H3", chol_resname), (" H3", 'CHO')),
              ((" C3", chol_resname), (" C3", 'CHO')),
              ((" C2", chol_resname), (" C2", 'CHO')),
              ((" C1", chol_resname), (" C1", 'CHO')),
              (("C10", chol_resname), ("C10", 'CHO')),
              (("C19", chol_resname), ("C19", 'CHO')),
              ((" C9", chol_resname), (" C9", 'CHO')),
              (("C11", chol_resname), ("C11", 'CHO')),
              (("C12", chol_resname), ("C12", 'CHO')),
              (("C13", chol_resname), ("C13", 'CHO')),
              (("C18", chol_resname), ("C18", 'CHO')),
              (("C17", chol_resname), ("C17", 'CHO')),
              (("C20", chol_resname), ("C20", 'CHO')),
              (("C22", chol_resname), ("C22", 'CHO')),
              (("C23", chol_resname), ("C23", 'CHO')),
              (("C24", chol_resname), ("C24", 'CHO')),
              (("C25", chol_resname), ("C25", 'CHO')),
              (("C27", chol_resname), ("C27", 'CHO')),
              (("C26", chol_resname), ("C26", 'CHO')),
              (("C21", chol_resname), ("C21", 'CHO')),
              (("C16", chol_resname), ("C16", 'CHO')),
              (("C15", chol_resname), ("C15", 'CHO')),
              (("C14", chol_resname), ("C14", 'CHO')),
              ((" C8", chol_resname), (" C8", 'CHO')),
              ((" C7", chol_resname), (" C7", 'CHO')),
              ((" C6", chol_resname), (" C6", 'CHO')),
              ((" H6", chol_resname), (" H6", 'CHO')),
              ((" C5", chol_resname), (" C5", 'CHO')),
              ((" C4", chol_resname), (" C4", 'CHO')),
              ((" O3", chol_resname), (" O3", 'CHO')),
              (("H3'", chol_resname), ("H3'", 'CHO'))]

    for swap in lookup:
        old_aname = swap[0][0].strip()
        old_resname = swap[0][1]
        new_aname = swap[1][0].strip()
        new_resname = swap[1][1]
        if old_aname == aname and \
           old_resname == resname:
            aname = new_aname
            resname = new_resname
            return aname, resname, True

    return aname, resname, False



from tautomer import Tautomer
import config

class Titsite:
    """Titrable Site with more than one Tautomer"""
    def __init__(self, res_number, molecule):
        """
        Args:
          res_number (int): number of the site in the .pdb
                            and .sites file
          molecule (TitratingMolecule): molecule to
                                        which the Site belongs

        # Site Details
        self._res_name (str): residue name
        self._type (str): site type ('c' or 'a')
                          'c' if site is cationic
                          'a' if site is anionic
        self._center (list) : Site focus center
            if dimension is 0, then center is defined in cent file
            elif dimension is 2, then center (z) is defined in cent file
                                 and center (x, y) are defined in boxsize variable
        self._center_original (list): Site geometric center on
                                      cent file when dimension is 2

        # Site Tautomers
        self._tautomers (dict): tautomers of Site (except reference tautomer)
                                key is tautomer name
                                value is tautomer instance
        self._ref_tautomer (Tautomer): reference tautomer instance

        # Site Atoms
        self._atoms (dict): existing atoms in site
                            key is atom id name
                            value is atom id number
        """
        self._molecule = molecule
        self._res_number = res_number
        self._res_name = ''
        self._atoms = {}
        self._tautomers = {}
        self._ref_tautomer = ''
        self._center = []
        self._type = ''

    # Set Methods
    def setTautomers(self, ntauts, resname):
        """Adds Tautomers from res_tauts to self._tautomers

        Args:
            res_tauts (list): tautomers from sites file
        """
        for i in range(ntauts):
            self._res_name = resname
            correct_number = str(i)
            tautomer = resname[:2] + correct_number

            tID = Tautomer(tautomer, self, self._molecule)
            self._tautomers[tautomer] = tID
            
    def addReferenceTautomer(self):
        """Gets last tautomer from .sites file adds one and saves it as the
        reference tautomer"""
        last_tautomer = max(sorted(self._tautomers.keys()))
        correct_number = str(int(last_tautomer[-1]) + 1)
        ref_tautomer = last_tautomer[:2] + correct_number
        tID = Tautomer(ref_tautomer, self, self._molecule)
        self._ref_tautomer = tID

    def addChargeSets(self):
        """Stores the charge set of each existing tautomer
        for the present Site"""
        for tautomer in list(self._tautomers.values()):
            tautomer.loadChargeSet(self._res_name,
                                   self._ref_tautomer)

    def addAtom(self, aname, anumb):
        """
        Args:
            aname (str): atom name
            anumb (int): atom id number
        """
        self._atoms[aname] = anumb

    def addCenter(self, center, boxsize=False, box_z=False):
        x = center[0]
        y = center[1]
        z = center[2]
        if boxsize:
            if config.params['pbc_dim'] != 2:
                raise Exception('ERROR: The original center is only '
                                'needed for 2-dimensional calculation')
            self._center_original = [x, y, z]
            x = boxsize * 10 / 2
            y = boxsize * 10 / 2
            z += box_z * 10
        self._center = [x, y, z]

    def addCenterH(self, center):
        self._centerH = center

    def setType(self, stype):
        self._type = stype

    # Get Methods
    def getName(self):
        return self._res_name

    def getTautomers(self):
        """Returns list of all tautomers instances
        except the tautomers of reference"""
        return list(self._tautomers.values())

    def getAtomNumbersList(self):
        return list(self._atoms.values())

    def getAtomNamesList(self):
        return list(self._atoms.keys())

    def getRefTautomerName(self):
        return self._ref_tautomer.getName()

    def getRefTautomer(self):
        return self._ref_tautomer

    def getMolecule(self):
        return self._molecule

    def getType(self):
        return self._type

    def getRefProtState(self):
        if self._type == 'c':
            reftau_prot_state = 1
        elif self._type == 'a':
            reftau_prot_state = -1
        return reftau_prot_state

    def getCenterOriginal(self):
        if config.params['pbc_dim'] != 2:
            raise Exception('ERROR: The original center is only '
                            'needed for 2-dimensional calculation')
        return self._center_original

    def getCenterH(self):
        return self._centerH

    def getCenter(self):
        return self._center

    def getResNumber(self):
        return self._res_number

    # Iter Methods
    def iterTautomers(self):
        for i in list(self._tautomers.values()):
            yield i

    def iterOrderedTautomersWithoutRef(self):
        for i in sorted(self._tautomers.keys()):
            yield self._tautomers[i]

    def getOrderedTautomersList(self):
        tmp = []
        for i in sorted(self._tautomers.keys()):
            tmp.append(self._tautomers[i])
        tmp.append(self._ref_tautomer)
        return tmp

    # Print Methods
    def __str__(self):
        tautomers = ' '.join(sorted(self._tautomers.keys()))
        natoms = len(self._atoms)
        center = [round(i, 2) for i in self._center]
        out = 'Site Number -> {0:5}   '\
              'Tautomers -> {1:30}  '\
              'Reference -> {2:5} '\
              'NAtoms -> {3:6} '\
              'Center -> {4}'.format(self._res_number, tautomers,
                                     self._ref_tautomer.getName(),
                                     natoms, center)
        return out



import config
import numpy as np
from formats import read_pdb_line, read_gro_line, new_pdb_line
from copy import copy
from log import reportWarning
from titsite import Titsite as Site
from tautomer import Tautomer
from concurrency import startPoolProcesses, runInteractionCalcs, runMCCalcs
from mc import MCrun

MAXNPKHALFS = 5

class Molecule(object):
    """Molecule with more than one titrable sites
    """
    def __init__(self):
        """
        # DelPhi Parameters
        self._delphi_refparams (list): input DelPhi parameters

        # Sites
        self._sites (dict): key is residue number
                            value is site object instance

        self._sites_order (list): values are site object instances
                                  order respects .sites

        # Atoms
        self._natoms (int): number of atoms in the molecule
        self._atoms (dict): key is atom id number
                            value is atom name
        self._atoms_array_position (dict):
            key is atom number
            value is atom position in pqr
            Example: if the first atom in pqr is number 541
                     then self._atoms_array_position[541] = 1

        # Offset of the first site
        This is only used because the membrane part is not finished
        and uses a pqr input of the first site
        self._InputPQRoffSet (list):
            x, y, z coordinates of the offset in the input pqr

        # Site Interactions
        self._site_interactions (list):
            contains tuples with pairs of site objects
        """
        self._delphi_refparams = ''
        self._sites = {}
        self._sites_order = []
        self._natoms = -1
        self._atoms = {}
        self._atoms_array_position = {}
        self._InputPQRoffSet = []
        self._site_interactions = []
        self.box = []
        self._dat_to_write = ''
        self._correct_names = {}
        self._correct_atoms = {}

        self._NTR = None
        self._CTR = None

        self.readTermini()

    def readTermini(self):
        NTR_atoms = []
        with open("{0}/{1}/sts/NTRtau1.st".format(config.script_dir,
                                                  config.params['ffID'])) as f:
            for line in f:
                parts = line.split()
                if len(parts) > 1:
                    NTR_atoms.append(parts[1].strip())

        CTR_atoms = []
        with open("{0}/{1}/sts/CTRtau1.st".format(config.script_dir,
                                                  config.params['ffID'])) as f:
            for line in f:
                parts = line.split()
                if len(parts) > 1:
                    CTR_atoms.append(parts[1].strip())

        self._NTR_atoms = NTR_atoms
        self._CTR_atoms = CTR_atoms

    # Set Methods
    def addAtom(self, aname, anumb, position):
        self._atoms[anumb] = aname
        self._atoms_array_position[anumb] = position

    def loadParams(self, params):
        self._delphi_refparams = params

    def loadDelPhiParams(self, delphimol):
        if delphimol != 'reload':
            self._delphimol = delphimol
        else:
            delphimol = self._delphimol
        self.p_atpos  = copy(delphimol.get_atpos())
        self.p_rad3   = copy(delphimol.get_rad3())
        self.p_chrgv4 = copy(delphimol.get_chrgv4())
        self.atinf    = copy(delphimol.get_atinf())
        self.p_iatmed = copy(delphimol.get_iatmed())

    # Get Methods
    def getSites(self):
        return self._sites

    def getSitesOrdered(self):
        return self._sites_order

    def getDelPhi(self):
        return self._delphimol

    def getArrayPosition(self, atom_id):
        """Returns the index of the atom in DelPhi position array
        Disclamer: used only in testing"""
        return self._atoms_array_position[atom_id]

    def getAtomsList(self):
        """Returns a list of atoms details
        (id, instance, atom index in DelPhi data structures)
        sorted by atom id number"""
        atom_list = []
        for atom in sorted(self._atoms.keys()):
            atom_list.append((self._atoms[atom], atom,
                              self._atoms_array_position[atom]))
        return atom_list

    def getDelPhiParams(self):
        return self._delphi_refparams

    def getTautomerNumber(self, t_index):
        """Uncodes the tautomer index t_index needed for multiprocessing
        and returns the corresponding tautomer instance"""
        i = -1
        for site in self._sites_order:
            i += 1
            if i == t_index:
                return site._ref_tautomer
            for tautomer in sorted(site._tautomers.values()):
                i += 1
                if i == t_index:
                    return tautomer
        raise Exception('Something is wrong')

    def getTautomerInstance(self, tautname, site_resnum):
        """Return the tautomer instance named tautname
        existent in the site of the residue number site_resnum"""
        site = self._sites[site_resnum]
        if tautname in list(site._tautomers.keys()):
            return site._tautomers[tautname]
        elif tautname == site._ref_tautomer._name:
            return site._ref_tautomer
        raise Exception('Something is very wrong!!!')

    def getNAtoms(self):
        """Return number of atoms of the Site"""
        return self._natoms

    def getTautNAtoms(self, taut_name):
        """Return number of atoms of the tautomer named taut_name
        Disclamer: it was used for testing"""
        for site in list(self._sites.values()):
            for tautomer in site.getTautomers():
                if tautomer._name == taut_name:
                    return tautomer._natoms

    def getPQROffset(self):
        return self._InputPQRoffSet

    # Iter Methods
    def iterAtoms(self):
        """Generator that iterates through all atoms details (name, id,
        position) in the Site"""
        for atom in sorted(self._atoms.keys()):
            yield (self._atoms[atom], atom,
                   self._atoms_array_position[atom])

    def iterAllSitesTautomers(self):
        """Generator that iterates through all Tautomer instances.
        The iteration is sorted by site and within each site the first
        to be yielded is the reference tautomer and then the rest of
        the tautomers by order"""
        for site in self._sites_order:
            yield site._ref_tautomer
            for tautomer in sorted(site._tautomers.values()):
                yield tautomer

    def iterNonRefSitesTautomers(self):
        """Generator that iterates through all Tautomer instances except
        the reference ones"""
        for site in self._sites_order:
            for tautomer in sorted(site._tautomers.values()):
                yield tautomer

    def iterAllSites(self):
        for site in self._sites_order:
            yield site

    # Printing Methods
    def printAllSites(self):
        """Prints all Site names"""
        for site in self._sites_order:
            print((site.getName()))

    def printAllTautomers(self):
        """Prints all Tautomer details"""
        for site in self._sites_order:
            print((site.getName()))
            for tautomer in site.iterTautomers():
                print(tautomer)
            print((site._ref_tautomer))

    # Input Files Manipulation Methods
    def makeSimpleSites(self):
        def SitesFileLine(resnumb, resname):
            for res in config.REGULARTITRATINGRES:
                if res[0:2] == resname[0:2]:
                    resname = res

            if resname in config.TITRABLETAUTOMERS:
                ntautomers = config.TITRABLETAUTOMERS[resname]
            else:
                for res in config.REGULARTITRATINGRES:
                    if res[0:2] == resname[0:2]:
                        ntautomers = config.TITRABLETAUTOMERS[res]
            text = '{0} '.format(resnumb)
            for i in range(ntautomers):
                text += '{0}tau{1} '.format(resname, i + 1)

            if resname in ('NTR', 'CTR'):
                resnumb += config.terminal_offset

            sID = self.addSite(resnumb)
            self.addTautomers(sID, ntautomers, resname)

            return text + '\n'

        def add2chain(chain, chain_res, resnumb, resname):
            if chain in chain_res:
                chain_res[chain][resnumb] = resname
            else:
                chain_res[chain] = {}
                chain_res[chain][resnumb] = resname

        sites_file = ''

        sites = []
        chain_res = {}
        with open(config.f_in) as f:
            nline = 0
            resnumb = None
            resname = None
            for line in f:
                nline += 1
                if 'ATOM ' == line[0:5]:
                    (aname, anumb, resname, chain,
                     resnumb, x, y, z) = read_pdb_line(line)
                    if chain == ' ':
                        chain = 'A'
                    if resnumb not in sites:
                        if len(sites) == 0 and 'NTR' not in sites and chain == 'A':
                            sites.append('NTR')
                            sites_file += SitesFileLine(resnumb, 'NTR')
                            self._NTR = resnumb
                        if resname in config.TITRABLERESIDUES and \
                           resname != 'NTR' and resname != 'CTR':
                            sites.append(resnumb)
                            sites_file += SitesFileLine(resnumb, resname)
                            add2chain(chain, chain_res, resnumb, resname)

                    if 'CTR' not in sites and \
                       aname in ('CT', 'OT', 'OT1', 'OT2', 'O1', 'O2', 'OXT') and chain == 'A':
                        sites.append('CTR')
                        sites_file += SitesFileLine(resnumb, 'CTR')
                        self._CTR = resnumb
                        add2chain(chain, chain_res, resnumb, 'CTR')
                        
        # Adding the reference tautomer to each site
        self.addReferenceTautomers()
        # Assigning a charge set to each tautomer
        self.addTautomersChargeSets()

        if config.debug:
            with open('tmp.sites', 'w') as f_new:
                f_new.write(sites_file)

        return chain_res, sites

    def makeSites(self, useTMPgro=None, sites=None):
        """Identifies titrable residues and checks integrity of the residue blocks
        (excluding Hydrogens)
        """
        def correctResName(resname):
            for res in config.REGULARTITRATINGRES:
                if res[0:2] == resname[0:2]:
                    return res
            return resname

        def makeSite(resnumb, resname):
            if resname in config.TITRABLETAUTOMERS:
                ntautomers = config.TITRABLETAUTOMERS[resname]
            else:
                for res in config.REGULARTITRATINGRES:
                    if res[0:2] == resname[0:2]:
                        ntautomers = config.TITRABLETAUTOMERS[res]
            sID = self.addSite(resnumb)
            self.addTautomers(sID, ntautomers, resname)

        def warning(resnumb, resname, res_atoms, mode=None):
            if mode == 'CYS' or resname == 'CYS':
                CYS_atoms = ['N', 'CA', 'CB', 'SG', 'C', 'O', 'H']
                if set(res_atoms).issubset(CYS_atoms) and \
                   set(CYS_atoms).issubset(res_atoms):
                    # no need to correct residue name
                    warn = '{0} {1} is assumed to be participating '\
                           'in a SS-bond'.format(resnumb, resname)
                    reportWarning(warn)
                    return

                CY0_atoms = ['N', 'CA', 'CB', 'SG', 'C', 'O', 'H', 'HG1']
                if set(res_atoms).issubset(CY0_atoms) and \
                   set(CY0_atoms).issubset(res_atoms):
                    self._correct_names[resnumb] = 'CY0'
                    return

                CY0_atoms = ['N', 'CA', 'CB', 'SG', 'C', 'O', 'H', 'HG']
                if set(res_atoms).issubset(CY0_atoms) and \
                   set(CY0_atoms).issubset(res_atoms):
                    self._correct_names[resnumb] = 'CY0'
                    self._correct_atoms[resnumb] = {'HG': 'HG1'}
                    return
                else:
                    warn = '{0} {1} failed integrity check'.format(resnumb,
                                                                   resname)
                    reportWarning(warn)

            elif resname not in config.TITRABLERESIDUES:
                return
            else:
                warn = '{0} {1} failed integrity check'.format(resnumb,
                                                               resname)
                reportWarning(warn)

        if config.f_in and not useTMPgro:
            filename = config.f_in
            filetype = 'pdb'
        else:
            filename = "TMP.gro"
            filetype = 'gro'

        resnumb = None
        cur_atoms = []
        prev_resnumb = None
        prev_resname = None
        with open(filename) as f:
            nline = 0
            maxnlines = 0
            for line in f:
                resname = None
                nline += 1
                if 'ATOM ' == line[0:5]:
                    (aname, anumb, resname, chain,
                     resnumb, x, y, z) = read_pdb_line(line)
                elif filetype == 'gro':
                    if nline > 2 and nline < maxnlines:
                        (aname, anumb, resname,
                         resnumb, x, y, z) = read_gro_line(line)
                    elif nline == 2:
                        natoms = int(line.strip())
                        maxnlines = natoms + 3

                if line == 'TER\n':
                    resnumb += 1
                if (prev_resnumb != resnumb or nline == maxnlines) and \
                   prev_resnumb:
                    if nline == maxnlines:
                        prev_resnumb = copy(resnumb)
                        resnumb = 'None'

                    if prev_resname in config.TITRABLERESIDUES or \
                       (prev_resnumb == self._NTR or resnumb == self._NTR) or \
                       (prev_resnumb == self._CTR or resnumb == self._CTR):
                        if prev_resnumb == self._NTR and resnumb != self._NTR:
                            prev_resname = correctResName(prev_resname)

                            if prev_resname in config.TITRABLERESIDUES:
                                res_tits = True
                            else:
                                res_tits = False

                            (integrity_nter,
                             integrity_site) = self.check_integrity(prev_resname,
                                                                    cur_atoms,
                                                                    nter=True,
                                                                    site=res_tits)
                            if integrity_nter:
                                nter_resnumb = prev_resnumb + config.terminal_offset
                                makeSite(nter_resnumb, 'NTR')
                                self._NTR = prev_resnumb
                            else:
                                warning(prev_resnumb, 'NTR', '')

                            if sites and prev_resnumb in sites:
                                if integrity_site:
                                    makeSite(prev_resnumb, prev_resname)
                                else:
                                    warning(prev_resnumb,
                                            prev_resname, cur_atoms)

                            prev_resnumb = None

                        # Dealing with the last residue and CTR
                        elif prev_resnumb == self._CTR and resnumb != self._CTR:
                            prev_resname = correctResName(prev_resname)

                            if prev_resname in config.TITRABLERESIDUES:
                                res_tits = True
                            else:
                                res_tits = False

                            integrity_cter, integrity_site = self.check_integrity(prev_resname,
                                                                                  cur_atoms,
                                                                                  cter=True,
                                                                                  site=res_tits)
                            if integrity_cter:
                                cter_resnumb = prev_resnumb + config.terminal_offset
                                makeSite(cter_resnumb, 'CTR')
                                self._CTR = prev_resnumb
                            else:
                                warning(prev_resnumb, 'CTR', '')

                            if sites and prev_resnumb in sites:
                                if integrity_site:
                                    makeSite(prev_resnumb, prev_resname)
                                else:
                                    warning(prev_resnumb, prev_resname, cur_atoms)

                        # Dealing with the previous residue
                        elif prev_resnumb:
                            if sites and prev_resnumb in sites:
                                prev_resname = correctResName(prev_resname)
                                res_atoms = copy(cur_atoms)
                                integrity_site = self.check_integrity(prev_resname,
                                                                      cur_atoms)
                                
                                if integrity_site:
                                    makeSite(prev_resnumb, prev_resname)
                                else:
                                    prev_resname = warning(prev_resnumb, prev_resname, res_atoms)
                            elif prev_resname == 'CYS':
                                # dealing with a CYS that is not in sites
                                res_atoms = copy(cur_atoms)
                                integrity_site = self.check_integrity(prev_resname,
                                                                      cur_atoms)

                                if not integrity_site:
                                    prev_resname = warning(prev_resnumb, prev_resname,
                                                           res_atoms, mode='CYS')
                    elif prev_resname == 'ALA':
                        #TODO: check residue block integrity for other non titrating residues
                        pass

                # Dealing with the new residue
                if prev_resnumb != resnumb:                    
                    cur_atoms = [aname]
                    prev_resnumb = resnumb
                    prev_resname = resname

                elif resnumb:
                    cur_atoms.append(aname)
                    if prev_resname in ('NTR', 'CTR') and \
                       prev_resname != resname:
                        prev_resname = resname

        # Adding the reference tautomer to each site
        self.addReferenceTautomers()
        # Assigning a charge set to each tautomer
        self.addTautomersChargeSets()

        # TODO: report blocks that failed the check (in .log file with
        # numbering reference to stepwise scheme)

        # TODO: add lipid residues

        if config.debug:
            print('exiting makeSites')

    def check_integrity(self, resname, res_atoms,
                        nter=False, cter=False, site=True):
        def read_anames(filename):
            res_atoms_st = []
            with open(filename) as f:
                for line in f.readlines()[1:]:
                    cols = line.strip().split()
                    aname = cols[1]
                    res_atoms_st.append(aname)
            return res_atoms_st

        def pop_atoms(anames1, anames2):
            trigger = False
            for aname in anames1:
                if aname not in anames2:
                    trigger = True
                    if config.debug:
                        print((aname, 'not in', resname))
                else:
                    anames2.remove(aname)
            if trigger:
                return anames2, False
            else:
                return anames2, True

        if config.debug:
            print('###### INTEGRITY CHECK ######')
            print(res_atoms)
        integrity = None

        if nter:
            st = '{0}/{1}/sts/NTRtau1.st'.format(config.script_dir,
                                                 config.params['ffID'])
            res_atoms_st = read_anames(st)
            res_atoms, integrity_nter = pop_atoms(res_atoms_st, res_atoms)
        elif cter:
            st = '{0}/{1}/sts/CTRtau1.st'.format(config.script_dir,
                                                 config.params['ffID'])
            res_atoms_st = read_anames(st)
            res_atoms, integrity_cter = pop_atoms(res_atoms_st, res_atoms)

        if site:
            main_chain = ('N', 'H', 'CA', 'C', 'O')
            if nter:
                main_chain = ()
            if cter:
                main_chain = ('N', 'H', 'CA')

            for aname in main_chain:
                if aname in res_atoms:
                    res_atoms.remove(aname)
                elif not nter and not cter:
                    integrity = False
            
            if config.debug:
                print(('i', integrity))

            if integrity is not False:
                filename = '{0}/{1}/sts/{2}tau1.st'.format(config.script_dir,
                                                           config.params['ffID'],
                                                           resname)
                res_atoms_st = read_anames(filename)
                res_atoms, integrity = pop_atoms(res_atoms_st, res_atoms)
                if config.debug:
                    print((res_atoms_st, res_atoms))
                    print(('i', integrity))

        if len(res_atoms) != 0 and integrity:
            raise Exception('Something is wrong')

        if nter:
            return integrity_nter, integrity
        elif cter:
            return integrity_cter, integrity
        else:
            return integrity

    def addSite(self, resnum):
        sID = Site(resnum, self)
        self._sites[resnum] = sID
        self._sites_order.append(sID)

        return sID

    def addTautomers(self, sID, ntautomers, resname):
        rootname = resname[0:2]
        sID._res_name = resname
        for itautomer in range(ntautomers):
            tautomer = rootname + str(itautomer)
            tID = Tautomer(tautomer, sID, sID._molecule)
            sID._tautomers[tautomer] = tID
        return sID._tautomers

    def loadSites(self, chains_length, chains_res):
        def siteHasNTautomers(chain, resnum):
            if resnum in chains_res[chain]:
                resname = chains_res[chain][resnum]
                if resname in config.TITRABLETAUTOMERS:
                    ntauts = config.TITRABLETAUTOMERS[resname]
                    return ntauts, resname
                else:
                    for res in config.TITRABLETAUTOMERS:
                        if res[:-1] == resname[:-1]:
                            ntauts = config.TITRABLETAUTOMERS[res]
                            return ntauts, res
            print(chains_length, chains_res, chain, resnum)
            print(type(resnum))
            print(resnum in chains_res[chain])
            print(resname in config.TITRABLETAUTOMERS)
            raise Exception('Something is wrong.')

        for chain in config.sites:
            sites = config.sites[chain]
            for site in sites:
                if 'N' in site:
                    resnum = int(site[:-1])
                    self._NTR = resnum
                    resnum += config.terminal_offset
                    res_ntauts = 3
                    res_name = 'NTR'
                elif 'C' in site:
                    resnum = int(site[:-1])
                    self._CTR = resnum
                    resnum += config.terminal_offset
                    res_ntauts = 4
                    res_name = 'CTR'
                else:
                    resnum = int(site)
                    res_ntauts, res_name = siteHasNTautomers(chain, resnum)

                sID = self.addSite(resnum)
                sID.setTautomers(res_ntauts, res_name)

        self.addReferenceTautomers()
        self.addTautomersChargeSets()


    def addReferenceTautomers(self):
        for site in list(self._sites.values()):
            site.addReferenceTautomer()

    def addTautomersChargeSets(self):
        for site in list(self._sites.values()):
            # reads .st files
            site.addChargeSets()

    def readIndexFile(self, f_ndx):
        protein_trigger = False
        protein_atoms = []
        with open(f_ndx) as f:
            for line in f:
                if "[ protein ]" in line.lower():
                    protein_trigger = True
                elif protein_trigger:
                    for atom in line.split():
                        protein_atoms.append(int(atom))
        return protein_atoms

    def deleteAllSites(self):
        for site in list(self._sites.values()):
            site._tautomers = {}
            site._ref_tautomer = ''
        self._sites = {}
        self._sites_order = []

    def readGROFile(self, groname):
        # TODO: For CpHMD read the index file:
        # self.readIndexFile()

        # Getting the "protein" with trjconv PBC mol
        # os.system('echo "Protein" | {GroDIR}/trjconv '
        #          '-f TMP_{config.sysname}.gro -s TMP_CpHMD.tpr '
        #          '-n TMP_CpHMD.ndx -o TMP_auxD1.gro '
        #          '-pbc mol -quiet'.format(GroDIR=paths['GroDIR'], sysname=config.sysname))

        new_pdb_content = """TITLE     Protein in water
REMARK    THIS IS A SIMULATION BOX
CRYST1  000.000  000.000  000.000  60.00  60.00  90.00 P 1           1
MODEL        1
"""
        new_pdb_content = ""
        site_positions = {}
        site_Hs = {}
        with open(groname) as f:
            line_counter = 0
            natoms_left = 0
            natoms = 0

            lines = f.read().splitlines()
            penul_line = lines[-2]
            last_line  = lines[-1]
            #final_res_numb = int(penul_line[:5].strip())
            #final_atom_numb = int(penul_line[15:20].strip())
            self.box = [float(i) for i in last_line.split()[:3]]

            if config.params['pbc_dim'] == 2:
                scaleP = (config.params['gsize'] - 1) / (self.box[0] * 10)
                scaleM = int(4 / scaleP + 0.5) * scaleP

                config.params['scaleP'] = scaleP
                config.params['scaleM'] = scaleM

            for line in lines:
                line_counter += 1
                if natoms_left > 0:

                    natoms_left -= 1
                    natoms += 1
                    aposition = natoms - 1

                    (aname, anumb, resname,
                     resnumb, x, y, z) = read_gro_line(line)
                    x, y, z = x * 10, y * 10, z * 10

                    #print resnumb, resname, aname, anumb, aposition, x, y, z

                    if resnumb == self._NTR and aname in self._NTR_atoms:
                        resname = 'NTR'
                        resnumb += config.terminal_offset
                    elif resnumb == self._CTR and aname in self._CTR_atoms:
                        resname = 'CTR'
                        resnumb += config.terminal_offset
                        if aname == 'C':
                            aname = 'CT'
                        #elif aname in ('O1', 'O2'):
                        #    aname = aname[0] + 'T' + aname[1]

                    if resnumb in list(self._correct_names.keys()):
                        resname = self._correct_names[resnumb]
                    if resnumb in list(self._correct_atoms.keys()) and \
                       aname in self._correct_atoms[resnumb]:
                        aname = self._correct_atoms[resnumb][aname]

                    self.addAtom(aname, anumb, aposition)
                    ref_tau_name = resname

                    if resnumb in list(self._sites.keys()) and \
                       aname in list(self._sites[resnumb]._ref_tautomer._charge_set.keys()):
                        #( aname not in ('N', 'H', 'C', 'O', 'CA') or 
                        #(aname in ('N', 'H', 'C', 'O', 'CA') and resname == 'NTR')):
                        # change res name to reference tautomer
                        ref_tau_name = self._sites[resnumb].getRefTautomerName()

                        # add atom to corresponding site
                        self._sites[resnumb].addAtom(aname, anumb)
                        if resnumb in site_positions:
                            site_positions[resnumb].append((x, y, z))
                            if aname[0] == 'H':
                                site_Hs[resnumb].append((x, y, z))
                        else:
                            site_positions[resnumb] = [(x, y, z)]
                            if aname[0] == 'H':
                                site_Hs[resnumb] = [(x, y, z)]
                            else:
                                site_Hs[resnumb] = []

                    new_pdb_content += new_pdb_line(aposition, aname,
                                                    ref_tau_name, resnumb,
                                                    x, y, z)

                elif line_counter == 2:
                    natoms_left = int(line.strip())
                    self._natoms = natoms_left

        new_pdb_content += 'TER\nENDMDL\n'
        with open('delphi_in_stmod.pdb', 'w') as f_new:
            f_new.write(new_pdb_content)

        # TODO: check Terminal_offset has to be bigger than the total number of residues
        # TODO: delete terminal_offset and use another approach to distinguish between N- and C-ter
        # TODO: check size xy > config.cutoff * 2
        # if so, raise Exception, and ask to change cutoff value

        # TODO: check if pbc_dim -> set gsizes from pdb size xy and ignore perfil

        for site in site_positions:
            if site in list(self._sites.keys()):
                pos_max = [-9999990, -999999, -999999]
                pos_min = [999999, 999999, 999999]
                focus_center = [0, 0, 0]
                for atom in site_positions[site]:
                    for i in range(3):
                        if pos_max[i] < atom[i]:
                            pos_max[i] = atom[i]
                        if pos_min[i] > atom[i]:
                            pos_min[i] = atom[i]
                focus_center[0] = (pos_max[0] + pos_min[0]) / 2
                focus_center[1] = (pos_max[1] + pos_min[1]) / 2
                focus_center[2] = (pos_max[2] + pos_min[2]) / 2
                if config.params['pbc_dim'] == 2:
                    self._sites[site].addCenter(focus_center,
                                                boxsize=self.box[0],
                                                box_z=self.box[2])
                else:
                    self._sites[site].addCenter(focus_center)
                hx, hy, hz = 0, 0, 0
                nHs = len(site_Hs[site])
                if nHs == 0:
                    sitename = self._sites[site].getName()
                    raise Exception('Site {1}{0} appears '
                                    'to have no Hydrogen atoms'.format(site,
                                                                       sitename))
                for h in site_Hs[site]:
                    hx += h[0]
                    hy += h[1]
                    hz += h[2]
                hx /= nHs
                hy /= nHs
                hz /= nHs
                Hcenter = (round(hx, 2), round(hy, 2), round(hz, 2))
                self._sites[site].addCenterH(Hcenter)

        if config.debug:
            with open('cent', 'w') as f_new:
                text = ''
                for site in site_positions:
                    if site in list(self._sites.keys()):
                        text += str(self._sites[site]._center) + '\n'
                f_new.write(text)
            if config.params['pbc_dim'] == 2:
                with open('centHs', 'w') as f_new:
                    text = ''
                    for site in site_positions:
                        if site in list(self._sites.keys()):
                            text += '{0} {1}\n'.format(site,
                                                       self._sites[site]._centerH)
                    f_new.write(text)
                with open('cent_original', 'w') as f_new:
                    text = ''
                    for site in site_positions:
                        if site in list(self._sites.keys()):
                            text += '{0} {1}\n'.format(site,
                                                       self._sites[site]._center_original)
                    f_new.write(text)

    def calcSiteInteractionsParallel(self, ncpus):
        """Calculates the pairwise interaction energies
        and writes them in a formatted .dat file
        Interactions are calculated using a pool of processes

        Args:
          ncpus (int): number of cpus to be used
        """
        if config.debug:
            self.writeDatHeader()
        counter = 0
        for site1 in self.getSitesOrdered()[:-1]:
            counter += 1
            for site2 in self.getSitesOrdered()[counter:]:
                self._site_interactions.append((site1, site2))

        sites = self.getSitesOrdered()
        self._npossible_states = [len(site._tautomers) + 1 for site in sites]

        self._interactions = []
        for nstate1 in range(sum(self._npossible_states)):
            self._interactions.append([])
            for nstate2 in range(sum(self._npossible_states)):
                self._interactions[nstate1].append(-999999)

        self._interactions_look = []
        aux = -1
        site = -1
        for nstates in self._npossible_states:
            site += 1
            self._interactions_look.append([])
            for state in range(nstates):
                aux += 1
                self._interactions_look[site].append(aux)

        ncpus = min(len(self._site_interactions), ncpus)
        if ncpus > 0:
            results = startPoolProcesses(runInteractionCalcs,
                                         self._site_interactions, ncpus,
                                         assign='ordered', merged_results=True)
        else:
            results = []
        to_write = ''
        T = float(config.params['temp'])
        for interaction in results:
            site1i = interaction[0]
            site2i = interaction[1]
            if self._interactions[site1i][site2i] != -999999:
                self._interactions[site1i][site2i] += interaction[2]
                self._interactions[site2i][site1i] += interaction[2]
                self._interactions[site1i][site2i] /= 2
                self._interactions[site2i][site1i] /= 2

                if config.debug:
                    col1 = interaction[3][6:18]
                    col2 = interaction[3][:6]
                    col3 = self._interactions[site1i][site2i] * (config.kBoltz * T)
                    to_write += '{0}{1} {2:13.6e}\n'.format(col1, col2, col3)
            else:
                self._interactions[site1i][site2i] = interaction[2]
                self._interactions[site2i][site1i] = interaction[2]

        if config.debug:
            with open('interactions.dat', 'a') as f_new:
                f_new.write(to_write)

    def calcInteractionNumber(self, inumber):
        """Calculates the pairwise interaction energies
        related to the two sites specified in inumber

        Args:
          inumber (int): site interaction pair code

        Ensures:
          to_write (str): site interaction energies formatted
        to be written in .dat file
        """
        sites = self._site_interactions[inumber]
        site1 = sites[0]
        site2 = sites[1]
        pairs = ((site1, site2), (site2, site1))

        iterAtomsList = self.getAtomsList()

        T = float(config.params['temp'])
        interactions = []
        for pair in pairs:
            site1 = pair[0]
            site2 = pair[1]
            ordered_tautomers_site1 = site1.getOrderedTautomersList()
            ordered_tautomers_site2 = site2.getOrderedTautomersList()
            for taut1 in ordered_tautomers_site1:
                site_atom_list = taut1._site.getAtomNumbersList()
                for taut2 in ordered_tautomers_site2:
                    interaction = taut1.calcInteractionWith(taut2,
                                                            site_atom_list,
                                                            iterAtomsList)

                    gg = interaction / (config.kBoltz * T)

                    nsite1 = self.getSitesOrdered().index(taut1._site)
                    nsite2 = self.getSitesOrdered().index(taut2._site)

                    state1 = int(taut1._name[-1])
                    state2 = int(taut2._name[-1])

                    if config.debug:
                        print((nsite1, nsite2, state1, state2,
                               taut1._name, taut2._name, gg))
                    site1i = self._interactions_look[nsite1][state1]
                    site2i = self._interactions_look[nsite2][state2]

                    datf = self.convertIntoDatFormat(taut1, taut2, interaction)
                    interactions.append((site1i, site2i, gg, datf))

        return interactions

    def convertIntoDatFormat(self, tau1, tau2, interaction):
        """Returns a .dat format interaction line

        Args:
          tau1 (Tautomer): first tautomer of the pair
          tau2 (Tautomer): second tautomer of the pair
          interaction (float): interaction energy of the tautomers pair

        Ensures:
          line (str): .dat formatted interaction line
        """
        site1_index = self._sites_order.index(tau1._site)
        tau1_index = tau1._name[-1]
        tau1_dat = '{0:>3} {1:>2}'.format(site1_index, tau1_index)
        site2_index = self._sites_order.index(tau2._site)
        tau2_index = tau2._name[-1]
        tau2_dat = '{0:>3} {1:>2}'.format(site2_index, tau2_index)
        line = '{0}   {1}   {2:13.6e}\n'.format(tau1_dat, tau2_dat, interaction)
        return line

    def writeDatHeader(self):
        """Writes pKint energies in .dat file header"""
        to_write = '{0}\n'.format(len(self._sites))
        for site in self._sites_order:
            to_write += '{0:3s}-{1:<7}{2:>2}  P  *\n'.format(site._res_name,
                                                             site._res_number,
                                                             len(site._tautomers) + 1)
            if site._type == 'c':
                tau_prot_state = 0
                ref_prot_state = 1
            elif site._type == 'a':
                tau_prot_state = 1
                ref_prot_state = 0

            for tautomer in site.iterOrderedTautomersWithoutRef():
                to_write += '{0:1d} {1:13.6e}\n'.format(tau_prot_state,
                                                        tautomer._dg)
            to_write += '{0:1d}  0.000000e+00\n'.format(ref_prot_state)
        if config.debug:
            with open('interactions.dat', 'w') as f_new:
                f_new.write(to_write)

    def calcpKint(self, unpacked_results):
        """Calculation the pKint of all tautomers
        """
        i = -1
        if config.debug:
            print('############ results ############')
        pkints = ''
        contributions = ''
        for tautomer in config.tit_mole.iterAllSitesTautomers():
            i += 1
            core_index = i % config.params['ncpus']
            job_index = int(i / config.params['ncpus'])
            result = unpacked_results[core_index][job_index]

            tautname    = result[0]
            tautresnumb = result[1]

            esolvationM = result[2]
            sitpotM     = result[3]
            esolvationS = result[4]
            sitpotS     = result[5]

            tautomer = config.tit_mole.getTautomerInstance(tautname,
                                                           tautresnumb)

            tautomer.saveDelPhiResults(esolvationS, sitpotS, esolvationM,
                                       sitpotM)
            if config.debug:
                print('### new tautomer ###')
                print((i, core_index, job_index, tautname,
                       tautomer._name, esolvationM, esolvationS))
                print((tautomer._name, tautomer._esolvationS,
                       len(tautomer._sitpotS), tautomer._esolvationM,
                       len(tautomer._sitpotM)))

            tautomer.calcBackEnergy()
            if not tautomer.isRefTautomer():
                tautomer.calcpKint()
                if config.debug:
                    print(('pkint', tautomer._name, tautomer._dg,
                           id(tautomer)))
                    pkints += '{} {} {}\n'.format(tautomer._name,
                                                  tautomer._dg, tautomer.pKint)
                    contributions += '{}{} {} {} {} {}\n'.format(tautomer._site._res_number,
                                                                 tautomer._name,
                                                                 tautomer.dG_solvationM,
                                                                 tautomer.dG_solvationS,
                                                                 tautomer.dG_solvationM - tautomer.dG_solvationS,
                                                                 tautomer.dG_back)
        if config.debug:
            with open('pkint', 'w') as f_new1, \
                 open('contributions', 'w') as f_new2:
                f_new1.write(pkints)
                f_new2.write(contributions)

    def calcpKhalfs(self, pH, nsites, avgs, pmean, pKs, count, mcsteps, pHmin, dpH):
        totalP = 0.0
        for site in range(nsites):
            mean = avgs[site] / float(mcsteps)
            totalP += mean
        
            p = pmean[site]
        
            if p > 0.5 and mean <= 0.5 or p < 0.5 and mean >= 0.5:
                pKhalf = pH - dpH * (mean - 0.5) / (mean - p)
                for i in range(MAXNPKHALFS): 
                    if pKs[site][i] == 100.0:
                        pKs[site][i] = pKhalf
                        break
        
            elif p > 1.5 and mean <= 1.5 or p < 1.5 and mean >= 1.5:
                pKhalf = pH - dpH * (mean - 1.5) / (mean - p)
                for i in range(MAXNPKHALFS): 
                    if pKs[site][i] == 100.0:
                        pKs[site][i] = pKhalf
                        break
        
            pmean[site] = mean
        return totalP, pKs, pmean

                
    def parallelMCrun(self, pHstep):
        sites = self.getSitesOrdered()
        nsites = len(sites)

        pHmin, pHmax = config.params['pHmin'], config.params['pHmax']
        dpH = config.params['pHstep']
        couple_min = config.params['couple_min']
        mcsteps = config.params['mcsteps']
        eqsteps = config.params['eqsteps']
        seed = config.params['seed']

        pHsteps = int(round(1 + (pHmax - pHmin) / dpH, 0))

        pH = pHmin + pHstep * dpH
        avgs, pmean, count = MCrun(nsites, self._npossible_states,
                                   self._possible_states_g,
                                   self._possible_states_occ,
                                   self._interactions,
                                   self._interactions_look,
                                   pHsteps, mcsteps, eqsteps, seed,
                                   pHmin, dpH, couple_min, pH)

        return (avgs, count)


    def runMC(self):
        def resize_list_of_lists(listn, maxsize, filler=None):
            for i in listn:
                diff = maxsize - len(i)
                for ii in range(diff):
                    i.append(filler)

        print('\nStart MC', end='\r')

        sites = self.getSitesOrdered()
        nsites = len(sites)
        self._possible_states     = [[] for site in sites]
        self._possible_states_g   = [[] for site in sites]
        self._possible_states_occ = [[] for site in sites]

        T = float(config.params['temp'])
        isite = -1
        for site in sites:
            isite += 1
            itaut = 0
            for tautomer in site.iterOrderedTautomersWithoutRef():
                dg = tautomer._dg / (config.kBoltz * T)
                self._possible_states_g[isite].append(dg)
                self._possible_states[isite].append(itaut)
                if site._type == 'c':
                    prot_state = 0
                elif site._type == 'a':
                    prot_state = 1

                self._possible_states_occ[isite].append(prot_state)
                itaut += 1

            if site._type == 'c':
                prot_state = 1
            elif site._type == 'a':
                prot_state = 0
            self._possible_states_occ[isite].append(prot_state)
            self._possible_states_g[isite].append(0.0)
        
        maxstates = max(self._npossible_states)
        resize_list_of_lists(self._possible_states, maxstates)
        resize_list_of_lists(self._possible_states_g, maxstates)
        resize_list_of_lists(self._possible_states_occ, maxstates, filler=-500)
        resize_list_of_lists(self._interactions_look, maxstates, filler=-500)

        pHmin, pHmax = config.params['pHmin'], config.params['pHmax']
        dpH = config.params['pHstep']
        couple_min = config.params['couple_min']
        mcsteps = config.params['mcsteps']
        eqsteps = config.params['eqsteps']
        seed = config.params['seed']

        pHsteps = int(round(1 + (pHmax - pHmin) / dpH, 0))

        pmeans_all = []
        counts_all = []
        avgs_all = []

        ncpus = min(config.params['ncpus'], nsites)
        results = startPoolProcesses(runMCCalcs,
                                     list(range(pHsteps)), ncpus,
                                     assign='ordered', merged_results=True)

        print('Ended MC')
        counter = 0
        for i in results:
            avgs_all.append(i[0])
            counts_all.append(i[1])

        pKs = np.array([[100.0 for ii in range(MAXNPKHALFS)] for i in range(nsites)])
        pmeans = avgs_all[0] / float(mcsteps)
        pmeans_raw = [100.0]
        for pHstep in range(1, pHsteps):
            pH = pHmin + pHstep * dpH
            totalP, pKs, pmeans = self.calcpKhalfs(pH, nsites, avgs_all[pHstep], pmeans, pKs, counts_all[pHstep], mcsteps, pHmin, dpH)
            pmeans_raw.append(totalP)
        pKas = pKs
        
        #pKas, pmeans_raw = mc.MCrun(nsites, self._npossible_states,
        #                            self._possible_states_g,
        #                            self._possible_states_occ,
        #                            self._interactions,
        #                            self._interactions_look,
        #                            pHsteps, mcsteps, eqsteps, seed,
        #                            pHmin, dpH, couple_min)

        print('\n\nResults')
        text_pks = ''
        text_prots = '# pH      total'
        c = -1
        for i in pKas:
            c += 1
            site = self.getSitesOrdered()[c]
            sitename = site.getName()
            if sitename in ('NTR', 'CTR'):
                resnumb = site.getResNumber() - config.terminal_offset
                text_prots += '     {0:3}'.format(sitename)
            else:
                resnumb = site.getResNumber()
                text_prots += '{0:5d}{1:3s}'.format(resnumb, sitename)
            text_pks += '{0} {1} {2}\n'.format(resnumb, sitename, i[0])
            if i[0] != 100.0:
                print(('{0} {1} {2}'.format(resnumb, sitename, round(i[0], 2))))
            else:
                print(('{0} {1} Not in range'.format(resnumb, sitename)))
        print('')

        if config.f_out:
            with open(config.f_out, 'w') as f:
                f.write(text_pks)

        pmeans = {}
        for pHstep in range(pHsteps):
            pH = pHmin + pHstep * dpH
            c = -1
            pmeans[pH] = {}
            text_prots += '\n{0:5.2f}'.format(pH)

            # total protonation
            #pmean = pmeans_raw[pHstep][c]
            pmean = pmeans_raw[pHstep]
            pmeans[pH]['total'] = pmean
            text_prots += '\t{0:7.4f}'.format(pmean)

            for site in self.getSitesOrdered():
                c += 1
                sitename = site.getName()
                sitenumb = site.getResNumber()
                #pmean = pmeans_raw[pHstep][c]
                pmean = avgs_all[pHstep][c]
                pmeans[pH][sitenumb] = pmean
                if pmean != 100.0:
                    text_prots += '\t{0:7.4f}'.format(pmean)
                else:
                    text_prots += '\t-'

        if config.f_prot_out:
            with open(config.f_prot_out, 'w') as f:
                f.write(text_prots)

        return pKas, pmeans



import config
import os
import sys


def reportToLOG(info):
    with open(config.f_log, 'a') as logfile:
        logfile.write(info + '\n')


def requiredParameterError(parameter):
    raise IOError('Required input parameter {0} '
                  'is not defined.'.format(parameter))


def inputVariableError(parameter, complaint, explanation):
    raise IOError('Input parameter {0} is not {1}\n '
                  '{2}'.format(parameter, complaint, explanation))


def reportWarning(info):
    warning = 'warning: {0}'.format(info)
    reportToLOG(warning)


def redirectOutput(mode, outputname):
    if mode == 'start':
        config.stdout = os.dup(sys.stdout.fileno())
        config.stdout_file = open(outputname, 'w')
        os.dup2(config.stdout_file.fileno(), sys.stdout.fileno())
    elif mode == 'stop':
        if not config.stdout:
            raise Exception('Output redirection has not been started, '
                            'thus it can not be ended.')
        sys.stdout.flush()
        os.dup2(config.stdout, sys.stdout.fileno())
        config.stdout_file.close()


def redirectErr(mode, outputname):
    if mode == 'start':
        config.stderr = os.dup(sys.stderr.fileno())
        config.stderr_file = open(outputname, 'w')
        os.dup2(config.stderr_file.fileno(), sys.stderr.fileno())
    elif mode == 'stop':
        if not config.stderr:
            raise Exception('Error redirection has not been started, '
                            'thus it can not be ended.')
        sys.stderr.flush()
        os.dup2(config.stderr, sys.stderr.fileno())
        config.stderr_file.close()


def checkDelPhiErrors(filename, mode=None):
    exceptions = []
    if mode == 'readFiles':
        exceptions = ['part of system outside the box!',
                      'has a net charge of']
    elif mode == 'runDelPhi':
        exceptions = ['part of system outside the box!']

    exit_trigger = False
    errors = ''
    with open(filename) as f:
        for line in f:
            ignore_error = False
            if 'WARNING' in line or 'Warning' in line:
                for exception in exceptions:
                    if exception in line:
                        ignore_error = True
                if not ignore_error:
                    exit_trigger = True
                    errors += line
    if exit_trigger:
        raise Exception('The following errors have been found on {0}: \n{1}'.format(filename, errors))
    if not config.debug and os.path.isfile(filename):
        os.remove(filename)



from copy import copy
import os

tit_mole = None
sites    = {}
pid      = None
debug    = False

input_conversion = {'grid_fill': 'perfil',
                    'pbc_dimensions': 'pbc_dim',
                    'convergence': 'maxc'}
stdout = None
stderr = None

stdout_file = None
stderr_file = None

njobs   = None # to be redeclared in concurrency
pb_time = None # to be redeclared in concurrency
total_jobs = None # to be redeclared in concurrency

# DelPhi related
params = {'perfil': 0.9,
          'gsize': 81,
          'scaleP': 1,
          'scaleM': 4,
          'precision': 'single',
          'ionicstr': 0.1,
          'bndcon': 3,
          'maxc': 0.01,
          'nlit': 500,
          'nonit': 0,
          'relfac': 0.75,
          'relpar': 0.75,
          'pbx': False,
          'pby': False,
          'ffID': 'G54A7',
          'ffinput': 'GROMOS',
          'pHmin': 0,
          'pHmax': 14,
          'pHstep': 0.25,
          'epssol': 80.0,
          'temp': 298.0,
          'seed': 1234567,
          'cutoff': 2.5,
          'pbc_dim': 0,
          'epsin': 20.0,
          'slice': 0.05,
          'ncpus': 1,
          'clean_pdb': True,
          'couple_min': 2.0,
          'mcsteps': 200000,
          'eqsteps': 1000}

# -1 NanoShaper off
# 0 connolly surface
# 1 skin
# 2 blobby
# 3 mesh
# 4 msms
nanoshaper = -1

# Paths
# TODO: include in dependencies
# in dependencies there is a pdb2pqr that has a important dat folder
fileDir = os.path.dirname(os.path.abspath(__file__))
script_dir = fileDir
pdb2pqr = "{0}/pdb2pqr/pdb2pqr.py".format(fileDir)
userff = "{0}/pdb2pqr/dat/GROMOS.DAT".format(fileDir)
usernames = "{0}/pdb2pqr/dat/GROMOS.names".format(fileDir)


# Input Files
f_in = None
f_in_extension = None
f_out = None
f_prot_out = None
f_log = "LOG"
f_dat = None

# Force Field Files
f_crg = None
f_siz = None

lipids = {'cholesterol': 'CHO',  # to edit
          'POPC': 'POP'}
lipid_residues = ['POX', 'PJ2', 'CHL']  # allowed residue names

# Constants
kBoltz = 5.98435e-6  # e^2/(Angstrom*K)
log10 = 2.302585092994046


# Tautomer Variables
terminal_offset = 2000

TITRABLETAUTOMERS = {'LYS': 3,
                     'HIS': 2,
                     'ASP': 4,
                     'GLU': 4,
                     'SER': 3,
                     'THR': 3,
                     'CYS': 3,
                     'CTR': 4,
                     'NTR': 3,
                     'TYR': 2}

TITRABLERESIDUES = list(TITRABLETAUTOMERS.keys())
REGULARTITRATINGRES = copy(list(TITRABLETAUTOMERS.keys()))

for res in REGULARTITRATINGRES:
    ntautomers = TITRABLETAUTOMERS[res]
    for i in range(ntautomers):
        TITRABLERESIDUES.append(res[0:2] + str(i + 1))



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pypka",
    version="0.0.6",
    author="Pedro Reis",
    author_email="pdreis@fc.ul.pt",
    description="A python module for flexible Poisson-Boltzmann based pKa calculations with proton tautomerism",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mms-fcul/PypKa",
    packages=setuptools.find_packages(),
    package_dir={'pypka': 'pypka'},
    package_data={'pypka': ['mc.so',
                            'addHtaut',
                            'G54A7/*',
                            'G54A7/sts/*',
                            'pdb2pqr/*',
                            'pdb2pqr/dat/*',
                            'pdb2pqr/extensions/*',
                            'pdb2pqr/src/*',
                            'pdb2pqr/ZSI/*',
                            'delphi4py/*',
                            'delphi4py/readFiles/*.so',
                            'delphi4py/rundelphi/*.so']},
    install_requires=['numpy', 'pytest', 'coverage'],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Cython",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3) ",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    entry_points={
        'console_scripts': [
            'pypka = pypka.pypka:CLI'
        ]
    }
)
